"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.
"""
import logging
from decimal import Decimal

from django.conf import settings

logger = logging.getLogger('Django Ledger Logger')
logger.setLevel(logging.INFO)

DJANGO_LEDGER_USE_CLOSING_ENTRIES = getattr(settings, 'DJANGO_LEDGER_USE_CLOSING_ENTRIES', True)
DJANGO_LEDGER_DEFAULT_CLOSING_ENTRY_CACHE_TIMEOUT = getattr(settings,
                                                            'DJANGO_LEDGER_DEFAULT_CLOSING_ENTRY_CACHE_TIMEOUT', 3600)
DJANGO_LEDGER_AUTHORIZED_SUPERUSER = getattr(settings, 'DJANGO_LEDGER_AUTHORIZED_SUPERUSER', False)
DJANGO_LEDGER_LOGIN_URL = getattr(settings, 'DJANGO_LEDGER_LOGIN_URL', settings.LOGIN_URL)
DJANGO_LEDGER_BILL_NUMBER_LENGTH = getattr(settings, 'DJANGO_LEDGER_BILL_NUMBER_LENGTH', 10)
DJANGO_LEDGER_INVOICE_NUMBER_LENGTH = getattr(settings, 'DJANGO_LEDGER_INVOICE_NUMBER_LENGTH', 10)
DJANGO_LEDGER_FORM_INPUT_CLASSES = getattr(settings, 'DJANGO_LEDGER_FORM_INPUT_CLASSES', 'input')
DJANGO_LEDGER_CURRENCY_SYMBOL = getattr(settings, 'DJANGO_LEDGER_CURRENCY_SYMBOL', '$')
DJANGO_LEDGER_SPACED_CURRENCY_SYMBOL = getattr(settings, 'DJANGO_LEDGER_SPACED_CURRENCY_SYMBOL', False)
DJANGO_LEDGER_SHOW_FEEDBACK_BUTTON = getattr(settings, 'DJANGO_LEDGER_SHOW_FEEDBACK_BUTTON', False)
DJANGO_LEDGER_FEEDBACK_EMAIL_LIST = getattr(settings, 'DJANGO_LEDGER_FEEDBACK_EMAIL_LIST', [])
DJANGO_LEDGER_FEEDBACK_FROM_EMAIL = getattr(settings, 'DJANGO_LEDGER_FEEDBACK_FROM_EMAIL', None)
DJANGO_LEDGER_VALIDATE_SCHEMAS_AT_RUNTIME = getattr(settings, 'DJANGO_LEDGER_VALIDATE_SCHEMAS_AT_RUNTIME', False)
DJANGO_LEDGER_TRANSACTION_MAX_TOLERANCE = getattr(settings, 'DJANGO_LEDGER_TRANSACTION_MAX_TOLERANCE', Decimal('0.02'))
DJANGO_LEDGER_TRANSACTION_CORRECTION = getattr(settings, 'DJANGO_LEDGER_TRANSACTION_CORRECTION', Decimal('0.01'))
DJANGO_LEDGER_ACCOUNT_CODE_GENERATE = getattr(settings, 'DJANGO_LEDGER_ACCOUNT_CODE_GENERATE', True)
DJANGO_LEDGER_ACCOUNT_CODE_GENERATE_LENGTH = getattr(settings, 'DJANGO_LEDGER_ACCOUNT_CODE_GENERATE_LENGTH', 5)
DJANGO_LEDGER_ACCOUNT_CODE_USE_PREFIX = getattr(settings, 'DJANGO_LEDGER_ACCOUNT_CODE_GENERATE_LENGTH', True)
DJANGO_LEDGER_JE_NUMBER_PREFIX = getattr(settings, 'DJANGO_LEDGER_JE_NUMBER_PREFIX', 'JE')
DJANGO_LEDGER_PO_NUMBER_PREFIX = getattr(settings, 'DJANGO_LEDGER_PO_NUMBER_PREFIX', 'PO')
DJANGO_LEDGER_ESTIMATE_NUMBER_PREFIX = getattr(settings, 'DJANGO_LEDGER_ESTIMATE_NUMBER_PREFIX', 'E')
DJANGO_LEDGER_INVOICE_NUMBER_PREFIX = getattr(settings, 'DJANGO_LEDGER_INVOICE_NUMBER_PREFIX', 'I')
DJANGO_LEDGER_BILL_NUMBER_PREFIX = getattr(settings, 'DJANGO_LEDGER_BILL_NUMBER_PREFIX', 'B')
DJANGO_LEDGER_VENDOR_NUMBER_PREFIX = getattr(settings, 'DJANGO_LEDGER_VENDOR_NUMBER_PREFIX', 'V')
DJANGO_LEDGER_CUSTOMER_NUMBER_PREFIX = getattr(settings, 'DJANGO_LEDGER_CUSTOMER_NUMBER_PREFIX', 'C')
DJANGO_LEDGER_EXPENSE_NUMBER_PREFIX = getattr(settings, 'DJANGO_LEDGER_EXPENSE_NUMBER_PREFIX', 'IEX')
DJANGO_LEDGER_INVENTORY_NUMBER_PREFIX = getattr(settings, 'DJANGO_LEDGER_INVENTORY_NUMBER_PREFIX', 'INV')
DJANGO_LEDGER_PRODUCT_NUMBER_PREFIX = getattr(settings, 'DJANGO_LEDGER_PRODUCT_NUMBER_PREFIX', 'IPR')
DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING = getattr(settings, 'DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING', 10)
DJANGO_LEDGER_JE_NUMBER_NO_UNIT_PREFIX = getattr(settings, 'DJANGO_LEDGER_JE_NUMBER_NO_UNIT_PREFIX', '000')

DJANGO_LEDGER_BILL_MODEL_ABSTRACT_CLASS = getattr(settings,
                                                  'DJANGO_LEDGER_BILL_MODEL_ABSTRACT_CLASS',
                                                  'django_ledger.models.bill.BillModelAbstract')

DJANGO_LEDGER_INVOICE_MODEL_ABSTRACT_CLASS = getattr(settings,
                                                     'DJANGO_LEDGER_INVOICE_MODEL_ABSTRACT_CLASS',
                                                     'django_ledger.models.invoice.InvoiceModelAbstract')

DJANGO_LEDGER_DEFAULT_COA = getattr(settings, 'DJANGO_LEDGER_DEFAULT_COA', None)

DJANGO_LEDGER_FINANCIAL_ANALYSIS = {
    'ratios': {
        'current_ratio': {
            'good_incremental': True,
            'ranges': {
                'healthy': 2.0,
                'watch': 1.5,
                'warning': 1.0,
                'critical': 0.5,
            }
        },
        'quick_ratio': {
            'good_incremental': True,
            'ranges': {
                'healthy': 1.5,
                'watch': 1.0,
                'warning': 0.5,
                'critical': 0.25
            }
        },
        'debt_to_equity': {
            'good_incremental': False,
            'ranges': {
                'healthy': 0.5,
                'watch': 1.0,
                'warning': 1.5,
                'critical': 2.0
            }
        },
        'return_on_equity': {
            'good_incremental': True,
            'ranges': {
                'healthy': 0.15,
                'watch': 0.10,
                'warning': 0.05,
                'critical': 0.00
            }
        },
        'return_on_assets': {
            'good_incremental': True,
            'ranges': {
                'healthy': 0.05,
                'watch': 0.03,
                'warning': 0.02,
                'critical': 0.01
            }
        },
        'net_profit_margin': {
            'good_incremental': True,
            'ranges': {
                'healthy': 0.10,
                'watch': 0.05,
                'warning': 0.03,
                'critical': 0.01
            }
        },
        'gross_profit_margin': {
            'good_incremental': True,
            'ranges': {
                'healthy': 0.50,
                'watch': 0.30,
                'warning': 0.20,
                'critical': 0.10
            }
        },
    }
}

DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR = getattr(settings, 'DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR', True)
DJANGO_LEDGER_THEME = getattr(settings, 'DJANGO_LEDGER_THEME', 'litera')

## MODEL ABSTRACTS ##

# DJANGO_LEDGER_ACCOUNT_MODEL = getattr(settings, 'DJANGO_LEDGER_ACCOUNT_MODEL', 'django_ledger.AccountModel')
# DJANGO_LEDGER_CHART_OF_ACCOUNTS_MODEL = getattr(settings, 'DJANGO_LEDGER_ACCOUNT_MODEL', 'django_ledger.ChartOfAccountModel')
# DJANGO_LEDGER_TRANSACTION_MODEL = getattr(settings, 'DJANGO_LEDGER_TRANSACTION_MODEL', 'django_ledger.TransactionModel')
# DJANGO_LEDGER_JOURNAL_ENTRY_MODEL = getattr(settings, 'DJANGO_LEDGER_JOURNAL_ENTRY_MODEL', 'django_ledger.JournalEntryModel')
# DJANGO_LEDGER_LEDGER_MODEL = getattr(settings, 'DJANGO_LEDGER_LEDGER_MODEL', 'django_ledger.LedgerModel')
# DJANGO_LEDGER_ENTITY_MODEL = getattr(settings, 'DJANGO_LEDGER_ENTITY_MODEL', 'django_ledger.EntityModel')
# DJANGO_LEDGER_ENTITY_STATE_MODEL = getattr(settings, 'DJANGO_LEDGER_ENTITY_STATE_MODEL', 'django_ledger.EntityStateModel')
# DJANGO_LEDGER_ENTITY_UNIT_MODEL = getattr(settings, 'DJANGO_LEDGER_ENTITY_UNIT_MODEL', 'django_ledger.EntityUnitModel')
# DJANGO_LEDGER_ESTIMATE_MODEL = getattr(settings, 'DJANGO_LEDGER_ESTIMATE_MODEL', 'django_ledger.EstimateModel')
# DJANGO_LEDGER_BILL_MODEL = getattr(settings, 'DJANGO_LEDGER_BILL_MODEL', 'django_ledger.BillModel')
# DJANGO_LEDGER_INVOICE_MODEL = getattr(settings, 'DJANGO_LEDGER_INVOICE_MODEL', 'django_ledger.InvoiceModel')
# DJANGO_LEDGER_PURCHASE_ORDER_MODEL = getattr(settings, 'DJANGO_LEDGER_PURCHASE_ORDER_MODEL', 'django_ledger.PurchaseOrderModel')
# DJANGO_LEDGER_CUSTOMER_MODEL = getattr(settings, 'DJANGO_LEDGER_CUSTOMER_MODEL', 'django_ledger.CustomerModel')
# DJANGO_LEDGER_VENDOR_MODEL = getattr(settings, 'DJANGO_LEDGER_VENDOR_MODEL', 'django_ledger.VendorModel')
# DJANGO_LEDGER_BANK_ACCOUNT_MODEL = getattr(settings, 'DJANGO_LEDGER_BANK_ACCOUNT_MODEL', 'django_ledger.BankAccountModel')
# DJANGO_LEDGER_CLOSING_ENTRY_MODEL = getattr(settings, 'DJANGO_LEDGER_CLOSING_ENTRY_MODEL', 'django_ledger.ClosingEntryModel')
# DJANGO_LEDGER_CLOSING_ENTRY_TRANSACTION_MODEL = getattr(settings, 'DJANGO_LEDGER_CLOSING_ENTRY_TRANSACTION_MODEL', 'django_ledger.ClosingEntryTransactionModel')
# DJANGO_LEDGER_UNIT_OF_MEASURE_MODEL = getattr(settings, 'DJANGO_LEDGER_UNIT_OF_MEASURE_MODEL', 'django_ledger.UnitOfMeasureModel')
# DJANGO_LEDGER_ITEM_TRANSACTION_MODEL = getattr(settings, 'DJANGO_LEDGER_ITEM_TRANSACTION_MODEL', 'django_ledger.ItemTransactionModel')
# DJANGO_LEDGER_ITEM_MODEL = getattr(settings, 'DJANGO_LEDGER_ITEM_MODEL', 'django_ledger.ItemModel')
# DJANGO_LEDGER_STAGED_TRANSACTION_MODEL = getattr(settings, 'DJANGO_LEDGER_STAGED_TRANSACTION_MODEL', 'django_ledger.StagedTransactionModel')
# DJANGO_LEDGER_IMPORT_JOB_MODEL = getattr(settings, 'DJANGO_LEDGER_IMPORT_JOB_MODEL', 'django_ledger.ImportJobModel')



# Generated by Django 5.0.3 on 2024-03-14 02:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0014_ledgermodel_ledger_xid_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='chartofaccountmodel',
            name='locked',
        ),
        migrations.AddField(
            model_name='chartofaccountmodel',
            name='active',
            field=models.BooleanField(default=True, verbose_name='Is Active'),
        ),
    ]



from django.urls import path

from django_ledger import views

urlpatterns = [
    path('<slug:entity_slug>/lastest/',
         views.InvoiceModelListView.as_view(),
         name='invoice-list'),
    path('<slug:entity_slug>/year/<int:year>/',
         views.InvoiceModelYearlyListView.as_view(),
         name='invoice-list-year'),
    path('<slug:entity_slug>/month/<int:year>/<int:month>/',
         views.InvoiceModelMonthlyListView.as_view(),
         name='invoice-list-month'),
    path('<slug:entity_slug>/create/',
         views.InvoiceModelCreateView.as_view(),
         name='invoice-create'),
    path('<slug:entity_slug>/create/estimate/<uuid:ce_pk>/',
         views.InvoiceModelCreateView.as_view(for_estimate=True),
         name='invoice-create-estimate'),
    path('<slug:entity_slug>/detail/<uuid:invoice_pk>/',
         views.InvoiceModelDetailView.as_view(),
         name='invoice-detail'),
    path('<slug:entity_slug>/update/<uuid:invoice_pk>/',
         views.InvoiceModelUpdateView.as_view(),
         name='invoice-update'),
    path('<slug:entity_slug>/update/<uuid:invoice_pk>/items/',
         views.InvoiceModelUpdateView.as_view(action_update_items=True),
         name='invoice-update-items'),
    path('<slug:entity_slug>/delete/<uuid:invoice_pk>/',
         views.InvoiceModelDeleteView.as_view(),
         name='invoice-delete'),

    # actions...
    path('<slug:entity_slug>/actions/<uuid:invoice_pk>/mark-as-draft/',
         views.InvoiceModelActionMarkAsDraftView.as_view(),
         name='invoice-action-mark-as-draft'),
    path('<slug:entity_slug>/actions/<uuid:invoice_pk>/mark-as-review/',
         views.InvoiceModelActionMarkAsReviewView.as_view(),
         name='invoice-action-mark-as-review'),
    path('<slug:entity_slug>/actions/<uuid:invoice_pk>/mark-as-approved/',
         views.InvoiceModelActionMarkAsApprovedView.as_view(),
         name='invoice-action-mark-as-approved'),
    path('<slug:entity_slug>/actions/<uuid:invoice_pk>/mark-as-paid/',
         views.InvoiceModelActionMarkAsPaidView.as_view(),
         name='invoice-action-mark-as-paid'),
    path('<slug:entity_slug>/actions/<uuid:invoice_pk>/mark-as-canceled/',
         views.InvoiceModelActionMarkAsCanceledView.as_view(),
         name='invoice-action-mark-as-canceled'),
    path('<slug:entity_slug>/actions/<uuid:invoice_pk>/mark-as-void/',
         views.InvoiceModelActionMarkAsVoidView.as_view(),
         name='invoice-action-mark-as-void'),

    path('<slug:entity_slug>/actions/<uuid:invoice_pk>/force-migrate/',
         views.InvoiceModelActionForceMigrateView.as_view(),
         name='invoice-action-force-migrate'),
    path('<slug:entity_slug>/actions/<uuid:invoice_pk>/lock/',
         views.InvoiceModelActionLockLedgerView.as_view(),
         name='invoice-action-lock-ledger'),
    path('<slug:entity_slug>/actions/<uuid:invoice_pk>/unlock/',
         views.InvoiceModelActionUnlockLedgerView.as_view(),
         name='invoice-action-unlock-ledger'),

]



# Generated by Django 4.2.2 on 2023-10-08 03:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0011_importjobmodel_alter_accountmodel_role_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='stagedtransactionmodel',
            name='activity',
            field=models.CharField(blank=True, choices=[('Operating', (('op', 'Operating'),)), ('Investing', (('inv_ppe', 'Purchase/Disposition of PPE'), ('inv_securities', 'Purchase/Disposition of Securities'), ('inv', 'Investing Activity Other'))), ('Financing', (('fin_std', 'Payoff of Short Term Debt'), ('fin_ltd', 'Payoff of Long Term Debt'), ('fin_equity', 'Issuance of Common Stock, Preferred Stock or Capital Contribution'), ('fin_dividends', 'Dividends or Distributions to Shareholders'), ('fin', 'Financing Activity Other')))], max_length=20, null=True, verbose_name='Proposed Activity'),
        ),
    ]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""
from random import randint
from typing import Optional

from django.forms import TextInput, Select, ModelForm, ChoiceField, ValidationError, CheckboxInput, HiddenInput
from django.utils.translation import gettext_lazy as _
from treebeard.forms import MoveNodeForm

from django_ledger.io import ACCOUNT_CHOICES_NO_ROOT
from django_ledger.models import ChartOfAccountModel
from django_ledger.models.accounts import AccountModel
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class AccountModelCreateForm(ModelForm):
    """
    A form for creating and managing account models within the system.

    Attributes
    ----------
    COA_MODEL : Model
        The Chart of Account Model being used in the form.
    """

    FORM_ID_SEP = '___'

    def __init__(self, coa_model: ChartOfAccountModel, *args, **kwargs):
        self.COA_MODEL: ChartOfAccountModel = coa_model
        super().__init__(*args, **kwargs)
        self.fields['role'].choices = ACCOUNT_CHOICES_NO_ROOT
        self.fields['code'].required = False
        self.fields['coa_model'].disabled = True
        self.fields['coa_model'].required = True

        self.form_id: str = self.get_form_id()

    def get_form_id(self) -> str:
        return f'account-model-create-form-{self.COA_MODEL.slug}{self.FORM_ID_SEP}{randint(100000, 999999)}'

    def clean_role_default(self):
        role_default = self.cleaned_data['role_default']
        if not role_default:
            return None
        return role_default

    def clean_coa_model(self):
        return self.COA_MODEL

    class Meta:
        model = AccountModel
        fields = [
            'code',
            'name',
            'role',
            'role_default',
            'balance_type',
            'active',
            'active',
            'coa_model'
        ]
        widgets = {
            'code': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Alpha Numeric (auto generated if not provided)...')
            }),
            'name': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Account Name...')
            }),
            'role': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'role_default': CheckboxInput(),
            'balance_type': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'coa_model': HiddenInput()
        }


class AccountModelUpdateForm(MoveNodeForm):
    """
    AccountModelUpdateForm

    A form for updating account model, inheriting from MoveNodeForm.

    Attributes
    ----------
    _position : ChoiceField
        A choice field for selecting the position.
    _ref_node_id : ChoiceField
        An optional choice field for selecting the relative node.
    """

    _position = ChoiceField(required=True,
                            label=_('Position'),
                            widget=Select(attrs={
                                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
                            }))
    _ref_node_id = ChoiceField(required=False,
                               label=_('Relative to'),
                               widget=Select(attrs={
                                   'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
                               }))


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['role'].disabled = True
        self.fields['coa_model'].disabled = True


    @classmethod
    def mk_dropdown_tree(cls, model, for_node: Optional[AccountModel] = None):
        """ Creates a tree-like list of choices """

        if not for_node:
            raise ValidationError(message='Must provide for_node argument.')

        qs = for_node.get_account_move_choice_queryset()

        return [
            (i.uuid, f'{"-" * (i.depth - 1)} {i}') for i in qs
        ]

    def clean_role(self):
        return self.instance.role

    def coa_model(self):
        return self.instance.coa_model

    def clean_role_default(self):
        role_default = self.cleaned_data['role_default']
        if not role_default:
            return None
        return role_default

    class Meta:
        model = AccountModel
        exclude = ('depth', 'numchild', 'path')
        widgets = {
            'role': HiddenInput(),
            'coa_model': HiddenInput(),
            'parent': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'balance_type': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'code': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'name': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'role_default': CheckboxInput(),
        }



from django.urls import path
from django_ledger import views

urlpatterns = [
    path('<slug:entity_slug>/list/current/',
         views.LedgerModelListView.as_view(show_current=True),
         name='ledger-list'),
    path('<slug:entity_slug>/list/visible/',
         views.LedgerModelListView.as_view(show_visible=True),
         name='ledger-list-visible'),
    path('<slug:entity_slug>/list/all/',
         views.LedgerModelListView.as_view(show_all=True),
         name='ledger-list-all'),
    path('<slug:entity_slug>/list/year/<int:year>/',
         views.LedgerModelYearListView.as_view(),
         name='ledger-list-year'),
    path('<slug:entity_slug>/list/month/<int:year>/<int:month>/',
         views.LedgerModelMonthListView.as_view(),
         name='ledger-list-month'),
    path('<slug:entity_slug>/create/',
         views.LedgerModelCreateView.as_view(),
         name='ledger-create'),
    path('<slug:entity_slug>/detail/<uuid:ledger_pk>/',
         views.LedgerModelUpdateView.as_view(),
         name='ledger-detail'),
    path('<slug:entity_slug>/update/<uuid:ledger_pk>/',
         views.LedgerModelUpdateView.as_view(),
         name='ledger-update'),
    path('<slug:entity_slug>/delete/<uuid:ledger_pk>/',
         views.LedgerModelDeleteView.as_view(),
         name='ledger-delete'),

    # ACTIONS...
    path('<slug:entity_slug>/action/<uuid:ledger_pk>/post/',
         views.LedgerModelModelActionView.as_view(action_name='post'),
         name='ledger-action-post'),
    path('<slug:entity_slug>/action/<uuid:ledger_pk>/post-journal-entries/',
         views.LedgerModelModelActionView.as_view(action_name='post_journal_entries'),
         name='ledger-action-post-journal-entries'),
    path('<slug:entity_slug>/action/<uuid:ledger_pk>/unpost/',
         views.LedgerModelModelActionView.as_view(action_name='unpost'),
         name='ledger-action-unpost'),
    path('<slug:entity_slug>/action/<uuid:ledger_pk>/lock/',
         views.LedgerModelModelActionView.as_view(action_name='lock'),
         name='ledger-action-lock'),
    path('<slug:entity_slug>/action/<uuid:ledger_pk>/lock-journal-entries/',
         views.LedgerModelModelActionView.as_view(action_name='lock_journal_entries'),
         name='ledger-action-lock-journal-entries'),
    path('<slug:entity_slug>/action/<uuid:ledger_pk>/unlock/',
         views.LedgerModelModelActionView.as_view(action_name='unlock'),
         name='ledger-action-unlock'),

    path('<slug:entity_slug>/action/<uuid:ledger_pk>/hide/',
         views.LedgerModelModelActionView.as_view(action_name='hide'),
         name='ledger-action-hide'),
    path('<slug:entity_slug>/action/<uuid:ledger_pk>/unhide/',
         views.LedgerModelModelActionView.as_view(action_name='unhide'),
         name='ledger-action-unhide'),

]



from django.urls import path

from django_ledger import views

urlpatterns = [
    path('<slug:entity_slug>/latest/',
         views.ClosingEntryModelListView.as_view(),
         name='closing-entry-list'),
    path('<slug:entity_slug>/list/year/<int:year>/',
         views.ClosingEntryModelYearListView.as_view(),
         name='closing-entry-list-year'),
    path('<slug:entity_slug>/list/month/<int:year>/<int:month>/',
         views.ClosingEntryModelMonthListView.as_view(),
         name='closing-entry-list-month'),
    path('<slug:entity_slug>/create/',
         views.ClosingEntryModelCreateView.as_view(),
         name='closing-entry-create'),
    path('<slug:entity_slug>/detail/<uuid:closing_entry_pk>/',
         views.ClosingEntryModelDetailView.as_view(),
         name='closing-entry-detail'),
    path('<slug:entity_slug>/update/<uuid:closing_entry_pk>/',
         views.ClosingEntryModelUpdateView.as_view(),
         name='closing-entry-update'),
    path('<slug:entity_slug>/delete/<uuid:closing_entry_pk>/',
         views.ClosingEntryDeleteView.as_view(),
         name='closing-entry-delete'),

    # ACTIONS...
    path('<slug:entity_slug>/action/<uuid:closing_entry_pk>/post/',
         views.ClosingEntryModelActionView.as_view(action_name='mark_as_posted'),
         name='closing-entry-action-mark-as-posted'),
    path('<slug:entity_slug>/action/<uuid:closing_entry_pk>/unpost/',
         views.ClosingEntryModelActionView.as_view(action_name='mark_as_unposted'),
         name='closing-entry-action-mark-as-unposted'),
    path('<slug:entity_slug>/action/<uuid:closing_entry_pk>/update-txs/',
         views.ClosingEntryModelActionView.as_view(action_name='update_transactions'),
         name='closing-entry-action-update-txs'),

]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.
"""

from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import ListView, UpdateView, CreateView, DetailView
from django.views.generic import RedirectView
from django.views.generic.detail import SingleObjectMixin

from django_ledger.forms.account import AccountModelUpdateForm, AccountModelCreateForm
from django_ledger.io.io_core import get_localdate
from django_ledger.models import EntityModel, ChartOfAccountModel
from django_ledger.models.accounts import AccountModel
from django_ledger.views.mixins import (
    YearlyReportMixIn, MonthlyReportMixIn, QuarterlyReportMixIn, DjangoLedgerSecurityMixIn,
    BaseDateNavigationUrlMixIn, EntityUnitMixIn, DateReportMixIn
)


class BaseAccountModelBaseView(DjangoLedgerSecurityMixIn):
    queryset = None
    coa_model = None

    def get_authorized_entity_queryset(self):
        qs = super().get_authorized_entity_queryset()
        return qs.select_related('admin', 'default_coa', 'default_coa__entity')

    def get_coa_model(self):
        if not self.coa_model:
            entity_model: EntityModel = self.get_authorized_entity_instance()
            self.coa_model = entity_model.chartofaccountmodel_set.get(slug__exact=self.kwargs['coa_slug'])
        return self.coa_model

    def get_queryset(self):
        if self.queryset is None:
            entity_model: EntityModel = self.get_authorized_entity_instance()
            coa_slug = self.kwargs['coa_slug']

            coa_model, account_model_qs = entity_model.get_coa_accounts(
                coa_model=entity_model.default_coa if coa_slug == entity_model.default_coa_slug else coa_slug,
                return_coa_model=True,
                active=False
            )

            account_model_qs = account_model_qs.select_related(
                'coa_model',
                'coa_model__entity'
            ).order_by(
                'coa_model', 'role', 'code'
            ).not_coa_root()

            self.coa_model = coa_model
            self.queryset = account_model_qs

        return super().get_queryset()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['coa_model'] = self.get_coa_model()
        return context


# Account Views ----
class AccountModelListView(BaseAccountModelBaseView, ListView):
    template_name = 'django_ledger/account/account_list.html'
    context_object_name = 'accounts'
    PAGE_TITLE = _('Entity Accounts')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE
    }
    active_only = False

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        coa_model: ChartOfAccountModel = self.get_coa_model()
        context['page_title'] = f'{coa_model.name} Accounts'
        context['header_title'] = f'{coa_model.name} Accounts'
        context['header_subtitle'] = self.get_authorized_entity_instance_name()
        context['header_subtitle_icon'] = 'ic:twotone-account-tree'
        return context

    def get_queryset(self):
        qs = super().get_queryset()
        if self.active_only:
            qs = qs.active()
        return qs



    def get(self, request, *args, **kwargs):
        response = super().get(request, *args, **kwargs)
        chart_of_accounts_model: ChartOfAccountModel = self.get_coa_model()
        if not chart_of_accounts_model.is_active():
            messages.error(request, _('WARNING: The chart of accounts list is inactive.'), extra_tags='is-danger')
        return response



class AccountModelCreateView(BaseAccountModelBaseView, CreateView):
    template_name = 'django_ledger/account/account_create.html'
    PAGE_TITLE = _('Create Account')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'ic:twotone-account-tree'

    }

    def get_form(self, form_class=None):
        return AccountModelCreateForm(
            coa_model=self.get_coa_model(),
            **self.get_form_kwargs()
        )

    def get_initial(self):
        return {
            'coa_model': self.get_coa_model(),
        }

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(*args, **kwargs)
        coa_model = self.get_coa_model()
        context['coa_model'] = coa_model
        context['header_subtitle'] = f'CoA: {coa_model.name}'
        return context

    def form_valid(self, form: AccountModelCreateForm):
        account_model: AccountModel = form.save(commit=False)
        coa_model = account_model.coa_model
        coa_model.insert_account(account_model=account_model)
        return HttpResponseRedirect(self.get_success_url())

    def get_success_url(self):
        coa_model: ChartOfAccountModel = self.get_coa_model()
        return coa_model.get_account_list_url()


class AccountModelUpdateView(BaseAccountModelBaseView, UpdateView):
    context_object_name = 'account'
    template_name = 'django_ledger/account/account_update.html'
    slug_url_kwarg = 'account_pk'
    slug_field = 'uuid'
    form_class = AccountModelUpdateForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['page_title'] = _('Update Account')
        context['header_title'] = _(f'Update Account: {self.object.code} - {self.object.name}')
        context['header_subtitle_icon'] = 'ic:twotone-account-tree'
        return context

    def get_success_url(self):
        coa_model: ChartOfAccountModel = self.get_coa_model()
        return coa_model.get_account_list_url()


class AccountModelDetailView(BaseAccountModelBaseView, RedirectView):

    def get_redirect_url(self, *args, **kwargs):
        loc_date = get_localdate()
        entity_model: EntityModel = self.get_authorized_entity_instance()
        return reverse('django_ledger:account-detail-month',
                       kwargs={
                           'entity_slug': entity_model.slug,
                           'account_pk': self.kwargs['account_pk'],
                           'coa_slug': self.kwargs['coa_slug'],
                           'year': loc_date.year,
                           'month': loc_date.month,
                       })


class AccountModelYearDetailView(BaseAccountModelBaseView,
                                 BaseDateNavigationUrlMixIn,
                                 EntityUnitMixIn,
                                 YearlyReportMixIn,
                                 DetailView):
    context_object_name = 'account'
    template_name = 'django_ledger/account/account_detail.html'
    slug_url_kwarg = 'account_pk'
    slug_field = 'uuid'
    DEFAULT_TXS_DAYS = 30
    extra_context = {
        'DEFAULT_TXS_DAYS': DEFAULT_TXS_DAYS,
        'header_subtitle_icon': 'ic:round-account-tree'
    }

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        account_model: AccountModel = context['object']
        context['header_title'] = f'Account {account_model.code} - {account_model.name}'
        context['page_title'] = f'Account {account_model.code} - {account_model.name}'
        txs_qs = account_model.transactionmodel_set.all().not_closing_entry().posted().order_by(
            'journal_entry__timestamp'
        ).select_related(
            'journal_entry',
            'journal_entry__entity_unit',
            'journal_entry__ledger__billmodel',
            'journal_entry__ledger__invoicemodel',
        )
        txs_qs = txs_qs.from_date(self.get_from_date())
        txs_qs = txs_qs.to_date(self.get_to_date())
        context['transactions'] = txs_qs
        return context


class AccountModelQuarterDetailView(AccountModelYearDetailView, QuarterlyReportMixIn):
    """
    Account Model Quarter Detail View
    """


class AccountModelMonthDetailView(AccountModelYearDetailView, MonthlyReportMixIn):
    """
    Account Model Month Detail View
    """


class AccountModelDateDetailView(AccountModelYearDetailView, DateReportMixIn):
    """
    Account Model Date Detail View
    """


# ACTIONS...
class BaseAccountModelActionView(BaseAccountModelBaseView,
                                 RedirectView,
                                 SingleObjectMixin):
    http_method_names = ['get']
    pk_url_kwarg = 'account_pk'
    action_name = None
    commit = True

    def get_redirect_url(self, *args, **kwargs):
        account_model: AccountModel = self.get_object()
        return account_model.get_coa_account_list_url()

    def get(self, request, *args, **kwargs):
        kwargs['user_model'] = self.request.user
        if not self.action_name:
            raise ImproperlyConfigured('View attribute action_name is required.')
        response = super(BaseAccountModelActionView, self).get(request, *args, **kwargs)
        account_model: AccountModel = self.get_object()

        try:
            getattr(account_model, self.action_name)(commit=self.commit, **kwargs)
        except ValidationError as e:
            messages.add_message(
                request,
                message=e.message,
                level=messages.ERROR,
                extra_tags='is-danger'
            )
        return response



# Generated by Django 4.1.5 on 2023-02-17 19:35

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0005_remove_itemmodel_django_ledg_invento_dbf206_idx_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='journalentrymodel',
            name='je_number',
            field=models.SlugField(editable=False, max_length=25, verbose_name='Journal Entry Number'),
        ),
        migrations.AlterField(
            model_name='journalentrymodel',
            name='timestamp',
            field=models.DateTimeField(default=django.utils.timezone.localtime, verbose_name='Timestamp'),
        ),
    ]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from django.contrib import messages
from django.core.exceptions import ValidationError, ImproperlyConfigured
from django.http import HttpResponseRedirect, HttpResponseForbidden
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import ArchiveIndexView, CreateView, DetailView, UpdateView, RedirectView
from django.views.generic.detail import SingleObjectMixin

from django_ledger.forms.estimate import (EstimateModelCreateForm, BaseEstimateModelUpdateForm,
                                          CanEditEstimateItemModelFormset, ReadOnlyEstimateItemModelFormset,
                                          DraftEstimateModelUpdateForm)
from django_ledger.models import EntityModel
from django_ledger.models.estimate import EstimateModel
from django_ledger.views import DjangoLedgerSecurityMixIn


class EstimateModelModelViewQuerySetMixIn:
    queryset = None

    def get_queryset(self):
        if self.queryset is None:
            entity_model: EntityModel = getattr(self, 'AUTHORIZED_ENTITY_MODEL')
            self.queryset = entity_model.estimatemodel_set.select_related('customer', 'entity')
        return super().get_queryset()


class EstimateModelListView(DjangoLedgerSecurityMixIn, EstimateModelModelViewQuerySetMixIn, ArchiveIndexView):
    template_name = 'django_ledger/estimate/estimate_list.html'
    context_object_name = 'estimate_list'
    PAGE_TITLE = _('Customer Estimates')
    date_field = 'created'
    paginate_by = 20
    paginate_orphans = 2
    allow_empty = True
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'eos-icons:job'
    }


class EstimateModelCreateView(DjangoLedgerSecurityMixIn, EstimateModelModelViewQuerySetMixIn, CreateView):
    PAGE_TITLE = _('Create Customer Estimate')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'eos-icons:job'
    }
    template_name = 'django_ledger/estimate/estimate_create.html'

    def get_form_class(self):
        return EstimateModelCreateForm

    def get_form(self, form_class=None):
        form_class = self.get_form_class()
        return form_class(entity_slug=self.kwargs['entity_slug'],
                          user_model=self.request.user,
                          **self.get_form_kwargs())

    def get_success_url(self):
        cj_model: EstimateModel = self.object
        return reverse('django_ledger:customer-estimate-detail',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug'],
                           'ce_pk': cj_model.uuid
                       })

    def form_valid(self, form):
        estimate_model: EstimateModel = form.save(commit=False)
        estimate_model.entity = self.AUTHORIZED_ENTITY_MODEL
        return super(EstimateModelCreateView, self).form_valid(form)


class EstimateModelDetailView(DjangoLedgerSecurityMixIn, EstimateModelModelViewQuerySetMixIn, DetailView):
    pk_url_kwarg = 'ce_pk'
    template_name = 'django_ledger/estimate/estimate_detail.html'
    PAGE_TITLE = _('Customer Estimate Detail')
    context_object_name = 'estimate_model'
    extra_context = {
        'hide_menu': True
    }
    http_method_names = ['get', 'post']

    def get_context_data(self, **kwargs):
        context = super(EstimateModelDetailView, self).get_context_data(**kwargs)
        ce_model: EstimateModel = self.object
        context['page_title'] = self.PAGE_TITLE
        context['header_title'] = self.PAGE_TITLE
        context['header_subtitle'] = ce_model.estimate_number
        context['header_subtitle_icon'] = 'eos-icons:job'
        context['estimate_item_list'] = ce_model.itemtransactionmodel_set.all()

        # PO Model Queryset...
        po_qs = ce_model.purchaseordermodel_set.for_entity(
            entity_model=self.kwargs['entity_slug']
        ) if ce_model.is_approved() else ce_model.purchaseordermodel_set.none()
        context['estimate_po_model_queryset'] = po_qs

        invoice_qs = ce_model.invoicemodel_set.for_entity(
            entity_model=self.kwargs['entity_slug']
        ) if ce_model.is_approved() else ce_model.invoicemodel_set.none()
        context['estimate_invoice_model_queryset'] = invoice_qs

        bill_qs = ce_model.billmodel_set.for_entity(
            entity_model=self.kwargs['entity_slug']
        ) if ce_model.is_approved() else ce_model.billmodel_set.none()
        context['estimate_bill_model_queryset'] = bill_qs

        if ce_model.is_contract():
            context['contract_progress'] = ce_model.get_contract_summary(
                po_qs=po_qs,
                invoice_qs=invoice_qs,
                bill_qs=bill_qs
            )

        return context

    def get_queryset(self):
        qs = super().get_queryset()
        return qs.prefetch_related('itemtransactionmodel_set')


class EstimateModelUpdateView(DjangoLedgerSecurityMixIn, EstimateModelModelViewQuerySetMixIn, UpdateView):
    template_name = 'django_ledger/estimate/estimate_update.html'
    pk_url_kwarg = 'ce_pk'
    context_object_name = 'estimate'
    PAGE_TITLE = _('Customer Estimate Update')
    http_method_names = ['get', 'post']

    action_update_items = False

    def get_form_class(self):
        estimate_model: EstimateModel = self.object
        if estimate_model.is_draft():
            return DraftEstimateModelUpdateForm
        return BaseEstimateModelUpdateForm

    def get_form(self, form_class=None):
        form_class = self.get_form_class()
        return form_class(
            entity_slug=self.kwargs['entity_slug'],
            user_model=self.request.user,
            **self.get_form_kwargs()
        )

    def get_context_data(self, itemtxs_formset=None, **kwargs):
        context = super(EstimateModelUpdateView, self).get_context_data(**kwargs)
        ce_model: EstimateModel = self.object
        context['page_title'] = self.PAGE_TITLE
        context['header_title'] = self.PAGE_TITLE
        context['header_subtitle'] = ce_model.title
        context['header_subtitle_icon'] = 'eos-icons:job'
        if not itemtxs_formset:
            itemtxs_qs, _ = ce_model.get_itemtxs_data()
            itemtxs_agg = itemtxs_qs.get_estimate_aggregate()
            if ce_model.can_update_items():
                itemtxs_formset = CanEditEstimateItemModelFormset(
                    entity_slug=self.kwargs['entity_slug'],
                    user_model=self.request.user,
                    customer_job_model=ce_model,
                    queryset=itemtxs_qs
                )
            else:
                itemtxs_formset = ReadOnlyEstimateItemModelFormset(
                    entity_slug=self.kwargs['entity_slug'],
                    user_model=self.request.user,
                    customer_job_model=ce_model,
                    queryset=itemtxs_qs
                )
        else:
            itemtxs_qs, _ = ce_model.get_itemtxs_data(queryset=itemtxs_formset.queryset)
            itemtxs_agg = itemtxs_qs.get_estimate_aggregate()

        context['ce_revenue_estimate__sum'] = itemtxs_agg['ce_revenue_estimate__sum']
        context['ce_cost_estimate__sum'] = itemtxs_agg['ce_cost_estimate__sum']
        context['itemtxs_qs'] = itemtxs_qs
        context['itemtxs_formset'] = itemtxs_formset
        return context

    def get_success_url(self):
        return reverse('django_ledger:customer-estimate-detail',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug'],
                           'ce_pk': self.kwargs['ce_pk']
                       })

    def get(self, request, entity_slug, ce_pk, *args, **kwargs):
        if self.action_update_items:
            return HttpResponseRedirect(
                redirect_to=reverse('django_ledger:customer-estimate-update',
                                    kwargs={
                                        'entity_slug': entity_slug,
                                        'ce_pk': ce_pk
                                    })
            )
        return super(EstimateModelUpdateView, self).get(request, entity_slug, ce_pk, *args, **kwargs)

    def post(self, request, entity_slug, ce_pk, *args, **kwargs):
        if self.action_update_items:

            if not request.user.is_authenticated:
                return HttpResponseForbidden()

            queryset = self.get_queryset()
            ce_model: EstimateModel = self.get_object(queryset=queryset)
            self.object = ce_model
            itemtxs_formset: CanEditEstimateItemModelFormset = CanEditEstimateItemModelFormset(request.POST,
                                                                                               user_model=self.request.user,
                                                                                               customer_job_model=ce_model,
                                                                                               entity_slug=entity_slug)
            if itemtxs_formset.has_changed():
                if itemtxs_formset.is_valid():
                    itemtxs_list = itemtxs_formset.save(commit=False)
                    entity_qs = EntityModel.objects.for_user(user_model=self.request.user)
                    entity_model: EntityModel = get_object_or_404(entity_qs, slug__exact=entity_slug)

                    for itemtxs in itemtxs_list:
                        itemtxs.ce_model_id = ce_model.uuid
                        itemtxs.clean()

                    itemtxs_list = itemtxs_formset.save()

                    ce_model.update_state()
                    ce_model.clean()
                    ce_model.save(update_fields=[
                        'revenue_estimate',
                        'labor_estimate',
                        'equipment_estimate',
                        'material_estimate',
                        'other_estimate',
                        'updated'
                    ])

                    messages.add_message(request,
                                         message=f'Customer estimate items saved.',
                                         level=messages.SUCCESS,
                                         extra_tags='is-success')

                    return self.render_to_response(context=self.get_context_data())
            context = self.get_context_data(itemtxs_formset=itemtxs_formset)
            return self.render_to_response(context=context)
        return super(EstimateModelUpdateView, self).post(request, *args, **kwargs)


# ---- ACTION VIEWS ----
class BaseEstimateActionView(DjangoLedgerSecurityMixIn,
                             EstimateModelModelViewQuerySetMixIn,
                             RedirectView,
                             SingleObjectMixin):
    http_method_names = ['get']
    pk_url_kwarg = 'ce_pk'
    action_name = None
    commit = True

    def get_redirect_url(self, entity_slug, ce_pk, *args, **kwargs):
        return reverse('django_ledger:customer-estimate-update',
                       kwargs={
                           'entity_slug': entity_slug,
                           'ce_pk': ce_pk
                       })

    def get(self, request, *args, **kwargs):
        if not self.action_name:
            raise ImproperlyConfigured('View attribute action_name is required.')
        response = super(BaseEstimateActionView, self).get(request, *args, **kwargs)
        ce_model: EstimateModel = self.get_object()

        try:
            getattr(ce_model, self.action_name)(commit=self.commit)
        except ValidationError as e:
            messages.add_message(request,
                                 message=e.message,
                                 level=messages.ERROR,
                                 extra_tags='is-danger')
        return response


class EstimateActionMarkAsDraftView(BaseEstimateActionView):
    action_name = 'mark_as_draft'


class EstimateActionMarkAsReviewView(BaseEstimateActionView):
    action_name = 'mark_as_review'


class EstimateActionMarkAsApprovedView(BaseEstimateActionView):
    action_name = 'mark_as_approved'


class EstimateActionMarkAsCompletedView(BaseEstimateActionView):
    action_name = 'mark_as_completed'


class EstimateActionMarkAsCanceledView(BaseEstimateActionView):
    action_name = 'mark_as_canceled'



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
Miguel Sanda <msanda@arrobalytics.com>
"""

from calendar import month_abbr
from random import randint
from typing import Union

from django import template
from django.core.exceptions import ValidationError
from django.db.models import Sum
from django.urls import reverse
from django.utils.formats import number_format

from django_ledger import __version__
from django_ledger.forms.app_filters import EntityFilterForm, ActivityFilterForm
from django_ledger.forms.feedback import BugReportForm, RequestNewFeatureForm
from django_ledger.io import ROLES_ORDER_ALL
from django_ledger.io.io_core import validate_activity, get_localdate
from django_ledger.models import BillModel, InvoiceModel, JournalEntryModel
from django_ledger.settings import (
    DJANGO_LEDGER_FINANCIAL_ANALYSIS,
    DJANGO_LEDGER_CURRENCY_SYMBOL,
    DJANGO_LEDGER_SPACED_CURRENCY_SYMBOL
)
from django_ledger.utils import get_default_entity_session_key, get_end_date_from_session

register = template.Library()


@register.simple_tag(name='current_version')
def current_version():
    return __version__


@register.simple_tag(name='currency_symbol')
def currency_symbol(spaced: bool = False):
    if spaced or DJANGO_LEDGER_SPACED_CURRENCY_SYMBOL:
        return f'{DJANGO_LEDGER_CURRENCY_SYMBOL} '
    return DJANGO_LEDGER_CURRENCY_SYMBOL


@register.filter(name='absolute')
def absolute(value):
    if value:
        if isinstance(value, str):
            value = float(value)
        return abs(value)


@register.filter(name='reverse_sign')
def reverse_sign(value):
    if value:
        if isinstance(value, str):
            value = float(value)
        return -value


@register.filter(name='currency_format')
def currency_format(value):
    if not value:
        value = 0.00
    return number_format(value, decimal_pos=2, use_l10n=True, force_grouping=True)


@register.filter(name='percentage')
def percentage(value):
    if value is not None:
        return '{0:,.2f}%'.format(value * 100)


@register.filter(name='last_four')
def last_four(value: str):
    if value:
        return '*' + value[-4:]
    return ''


@register.inclusion_tag('django_ledger/components/icon.html')
def icon(icon_name, size):
    return {
        'icon': icon_name,
        'size': size
    }


@register.inclusion_tag('django_ledger/financial_statements/tags/balance_sheet_statement.html', takes_context=True)
def balance_sheet_statement(context, io_model, to_date=None):
    user_model = context['user']
    activity = context['request'].GET.get('activity')
    entity_slug = context['view'].kwargs.get('entity_slug')

    if not to_date:
        to_date = context['to_date']

    io_digest = io_model.digest(
        activity=activity,
        user_model=user_model,
        equity_only=False,
        entity_slug=entity_slug,
        unit_slug=context['unit_slug'],
        by_unit=context['by_unit'],
        to_date=to_date,
        signs=True,
        process_groups=True,
        balance_sheet_statement=True)

    return {
        'entity_slug': entity_slug,
        'user_model': user_model,
        'tx_digest': io_digest.get_io_data(),
    }


@register.inclusion_tag('django_ledger/financial_statements/tags/cash_flow_statement.html', takes_context=True)
def cash_flow_statement(context, io_model):
    user_model = context['user']
    entity_slug = context['view'].kwargs.get('entity_slug')
    from_date = context['from_date']
    to_date = context['to_date']

    io_digest = io_model.digest(
        cash_flow_statement=True,
        by_activity=True,
        user_model=user_model,
        equity_only=False,
        signs=True,
        entity_slug=entity_slug,
        unit_slug=context['unit_slug'],
        by_unit=context['by_unit'],
        from_date=from_date,
        to_date=to_date,
        process_groups=True)

    return {
        'entity_slug': entity_slug,
        'user_model': user_model,
        'tx_digest': io_digest.get_io_data()
    }


@register.inclusion_tag('django_ledger/financial_statements/tags/income_statement.html', takes_context=True)
def income_statement_table(context, io_model, from_date=None, to_date=None):
    user_model = context['user']
    activity = context['request'].GET.get('activity')
    activity = validate_activity(activity, raise_404=True)
    entity_slug = context['view'].kwargs.get('entity_slug')

    if not from_date:
        from_date = context['from_date']
    if not to_date:
        to_date = context['to_date']

    io_digest = io_model.digest(
        activity=activity,
        user_model=user_model,
        entity_slug=entity_slug,
        unit_slug=context['unit_slug'],
        by_unit=context['by_unit'],
        from_date=from_date,
        to_date=to_date,
        equity_only=True,
        process_groups=True,
        income_statement=True,
        signs=True
    )

    return {
        'entity_slug': entity_slug,
        'user_model': user_model,
        'tx_digest': io_digest.get_io_data()
    }


@register.inclusion_tag('django_ledger/bank_account/tags/bank_accounts_table.html', takes_context=True)
def bank_account_table(context, bank_account_qs):
    entity_slug = context['view'].kwargs['entity_slug']
    return {
        'bank_account_qs': bank_account_qs,
        'entity_slug': entity_slug
    }


@register.inclusion_tag('django_ledger/data_import/tags/data_import_job_list_table.html', takes_context=True)
def data_import_job_list_table(context):
    return context


@register.inclusion_tag('django_ledger/data_import/tags/data_import_job_txs_table.html', takes_context=True)
def data_import_job_txs_pending(context, staged_txs_formset):
    return {
        'entity_slug': staged_txs_formset.IMPORT_JOB_MODEL.entity_slug,
        'import_job_model': staged_txs_formset.IMPORT_JOB_MODEL,
        'staged_txs_formset': staged_txs_formset
    }


@register.inclusion_tag('django_ledger/data_import/tags/data_import_job_txs_imported.html', takes_context=True)
def data_import_job_txs_imported(context, import_job_model):
    return {
        'entity_slug': import_job_model.entity_slug,
        'import_job_model': import_job_model,
        'imported_txs': import_job_model.stagedtransactionmodel_set.all().is_imported()
    }


@register.inclusion_tag('django_ledger/journal_entry/tags/je_table.html', takes_context=True)
def jes_table(context, journal_entry_qs, next_url=None):
    entity_slug = context['view'].kwargs['entity_slug']
    ledger_pk = context['view'].kwargs['ledger_pk']
    if not next_url:
        next_url = reverse('django_ledger:je-list',
                           kwargs={
                               'entity_slug': entity_slug,
                               'ledger_pk': ledger_pk
                           })
    return {
        'journal_entry_qs': journal_entry_qs,
        'entity_slug': entity_slug,
        'ledger_pk': ledger_pk,
        'next_url': next_url
    }


@register.inclusion_tag('django_ledger/transactions/tags/txs_table.html')
def transactions_table(object_type: Union[JournalEntryModel, BillModel, InvoiceModel], style='detail'):
    if isinstance(object_type, JournalEntryModel):
        transaction_model_qs = object_type.transactionmodel_set.all().with_annotated_details().order_by(
            '-timestamp')
    elif isinstance(object_type, BillModel):
        transaction_model_qs = object_type.get_transaction_queryset(annotated=True).order_by('-timestamp')
    elif isinstance(object_type, InvoiceModel):
        transaction_model_qs = object_type.get_transaction_queryset(annotated=True).order_by('-timestamp')
    else:
        raise ValidationError(
            'Cannot handle object of type {} to get transaction model queryset'.format(type(object_type))
        )

    total_credits = sum(tx.amount for tx in transaction_model_qs if tx.is_credit())
    total_debits = sum(tx.amount for tx in transaction_model_qs if tx.is_debit())

    return {
        'style': style,
        'transaction_model_qs': transaction_model_qs,
        'total_debits': total_debits,
        'total_credits': total_credits,
        'object': object_type
    }


@register.inclusion_tag('django_ledger/ledger/tags/ledgers_table.html', takes_context=True)
def ledgers_table(context, ledger_model_qs):
    return {
        'ledger_model_qs': ledger_model_qs,
        'entity_slug': context['view'].kwargs['entity_slug'],
    }


@register.inclusion_tag('django_ledger/invoice/tags/invoice_table.html', takes_context=True)
def invoice_table(context, invoice_qs):
    return {
        'invoices': invoice_qs,
        'entity_slug': context['view'].kwargs['entity_slug']
    }


@register.inclusion_tag('django_ledger/bills/tags/bill_table.html', takes_context=True)
def bill_table(context, bill_qs):
    return {
        'bills': bill_qs,
        'entity_slug': context['view'].kwargs['entity_slug']
    }


@register.inclusion_tag('django_ledger/closing_entry/tags/closing_entry_table.html', takes_context=True)
def closing_entry_table(context, closing_entry_qs):
    return {
        'closing_entry_list': closing_entry_qs,
        'entity_slug': context['view'].kwargs['entity_slug']
    }


@register.inclusion_tag('django_ledger/closing_entry/tags/closing_entry_txs_table.html', takes_context=True)
def closing_entry_txs_table(context, closing_entry_txs_qs):
    ce_txs_list = list(closing_entry_txs_qs)
    ce_txs_list.sort(key=lambda ce_txs: ROLES_ORDER_ALL.index(ce_txs.account_model.role))
    return {
        'ce_txs_list': ce_txs_list,
        'entity_slug': context['view'].kwargs['entity_slug']
    }


@register.inclusion_tag('django_ledger/purchase_order/includes/po_table.html', takes_context=True)
def po_table(context, purchase_order_qs):
    return {
        'po_list': purchase_order_qs,
        'entity_slug': context['view'].kwargs['entity_slug']
    }


@register.inclusion_tag('django_ledger/account/tags/accounts_table.html', takes_context=True)
def accounts_table(context, accounts_qs, title=None):
    return {
        'title': title,
        'entity_slug': context['view'].kwargs['entity_slug'],
        'accounts_gb': accounts_qs.gb_bs_role(),
    }


@register.inclusion_tag('django_ledger/customer/tags/customer_table.html', takes_context=True)
def customer_table(context):
    return context


@register.inclusion_tag('django_ledger/vendor/tags/vendor_table.html', takes_context=True)
def vendor_table(context):
    return context


@register.inclusion_tag('django_ledger/account/tags/account_txs_table.html', takes_context=True)
def account_txs_table(context, txs_qs):
    return {
        'transactions': txs_qs,
        'total_credits': sum(tx.amount for tx in txs_qs if tx.tx_type == 'credit'),
        'total_debits': sum(tx.amount for tx in txs_qs if tx.tx_type == 'debit'),
        'entity_slug': context['view'].kwargs['entity_slug'],
        'account_pk': context['view'].kwargs['account_pk']
    }


@register.inclusion_tag('django_ledger/components/breadcrumbs.html', takes_context=True)
def nav_breadcrumbs(context):
    entity_slug = context['view'].kwargs.get('entity_slug')
    coa_slug = context['view'].kwargs.get('coa_slug')
    ledger_pk = context['view'].kwargs.get('entity_slug')
    account_pk = context['view'].kwargs.get('account_pk')
    return {
        'entity_slug': entity_slug,
        'coa_slug': coa_slug,
        'ledger_pk': ledger_pk,
        'account_pk': account_pk
    }


@register.inclusion_tag('django_ledger/components/default_entity.html', takes_context=True)
def default_entity(context):
    user = context['user']
    session_key = get_default_entity_session_key()
    session = context['request'].session
    session_entity_data = session.get(session_key)
    identity = randint(0, 1000000)
    try:
        entity_uuid = session_entity_data['entity_uuid']
        default_entity_form = EntityFilterForm(
            user_model=user,
            form_id=identity,
            current_entity_uuid=entity_uuid
        )
    except TypeError or KeyError:
        default_entity_form = EntityFilterForm(
            user_model=user,
            form_id=identity,
        )

    return {
        'default_entity_form': default_entity_form,
        'form_id': identity,
    }


@register.simple_tag(takes_context=True)
def session_entity_name(context, request=None):
    session_key = get_default_entity_session_key()
    if not request:
        request = context.get('request')
    try:
        session = request.session
        entity_name = session.get(session_key)['entity_name']
    except AttributeError:
        entity_name = 'Django Ledger'
    except KeyError:
        entity_name = 'Django Ledger'
    except TypeError:
        entity_name = 'Django Ledger'
    return entity_name


# todo: rename template to activity_form_filter.
@register.inclusion_tag('django_ledger/components/activity_form.html', takes_context=True)
def activity_filter(context):
    request = context['request']
    activity = request.GET.get('activity')
    if activity:
        activity_form = ActivityFilterForm(initial={
            'activity': activity
        })
    else:
        activity_form = ActivityFilterForm()

    return {
        'activity_form': activity_form,
        'form_path': context['request'].path
    }


@register.inclusion_tag('django_ledger/components/date_picker.html', takes_context=True)
def date_picker(context, nav_url=None, date_picker_id=None):
    try:
        entity_slug = context['view'].kwargs.get('entity_slug')
    except KeyError:
        entity_slug = context['entity_slug']

    if not date_picker_id:
        date_picker_id = f'djl-datepicker-{randint(10000, 99999)}'

    if 'date_picker_ids' not in context:
        context['date_picker_ids'] = list()
    context['date_picker_ids'].append(date_picker_id)

    date_navigation_url = nav_url if nav_url else context.get('date_navigation_url')
    return {
        'entity_slug': entity_slug,
        'date_picker_id': date_picker_id,
        'date_navigation_url': date_navigation_url
    }


@register.simple_tag(takes_context=True)
def get_current_end_date_filter(context):
    entity_slug = context['view'].kwargs.get('entity_slug')
    return get_end_date_from_session(entity_slug, context['request'])


@register.inclusion_tag('django_ledger/components/chart_container.html')
def chart_container(chart_id, endpoint=None):
    return {
        'chart_id': chart_id,
        'endpoint': endpoint
    }


@register.inclusion_tag('django_ledger/components/modals.html', takes_context=True)
def modal_action(context, model, http_method: str = 'post', entity_slug: str = None):
    if not entity_slug:
        entity_slug = context['view'].kwargs['entity_slug']
    action_url = model.get_mark_as_paid_url(entity_slug=entity_slug)
    return {
        'object': model,
        'action_url': action_url,
        'http_method': http_method,
        'message': f'Do you want to mark {model.__class__._meta.verbose_name} {model.get_document_id()} as paid?'
    }


@register.inclusion_tag('django_ledger/components/modals_v2.html', takes_context=True)
def modal_action_v2(context, model, action_url: str, message: str, html_id: str, http_method: str = 'get'):
    return {
        'object': model,
        'action_url': action_url,
        'http_method': http_method,
        'message': message,
        'html_id': html_id
    }


@register.simple_tag
def fin_ratio_max_value(ratio: str):
    params = DJANGO_LEDGER_FINANCIAL_ANALYSIS['ratios'][ratio]['ranges']
    return params['healthy']


@register.filter
def fin_ratio_threshold_class(value, ratio):
    if value:
        params = DJANGO_LEDGER_FINANCIAL_ANALYSIS['ratios'][ratio]
        ranges = params['ranges']

        if params['good_incremental']:
            if value <= ranges['critical']:
                return 'is-danger'
            elif value <= ranges['warning']:
                return 'is-warning'
            elif value <= ranges['watch']:
                return 'is-primary'
            return 'is-success'
        else:
            if value >= ranges['critical']:
                return 'is-danger'
            elif value >= ranges['warning']:
                return 'is-warning'
            elif value >= ranges['watch']:
                return 'is-primary'
            return 'is-success'


@register.inclusion_tag('django_ledger/components/feedback_button.html', takes_context=True)
def feedback_button(context, button_size_class: str = 'is-small', color_class: str = 'is-success', icon_id: str = None):
    bug_modal_html_id = f'djl-bug-button-{randint(10000, 99999)}'
    feature_modal_html_id = f'djl-feature-button-{randint(10000, 99999)}'
    bug_form = BugReportForm()
    feature_form = RequestNewFeatureForm()
    next_url = context['request'].path
    return {
        'icon_id': icon_id,
        'bug_modal_html_id': bug_modal_html_id,
        'feature_modal_html_id': feature_modal_html_id,
        'button_size_class': button_size_class,
        'color_class': color_class,
        'bug_form': bug_form,
        'feature_form': feature_form,
        'next_url': next_url
    }


@register.inclusion_tag('django_ledger/components/period_navigator.html', takes_context=True)
def period_navigation(context, base_url: str):
    kwargs = dict()
    entity_slug = context['view'].kwargs['entity_slug']
    kwargs['entity_slug'] = entity_slug

    if context['view'].kwargs.get('ledger_pk'):
        kwargs['ledger_pk'] = context['view'].kwargs.get('ledger_pk')

    if context['view'].kwargs.get('account_pk'):
        kwargs['account_pk'] = context['view'].kwargs.get('account_pk')

    if context['view'].kwargs.get('unit_slug'):
        kwargs['unit_slug'] = context['view'].kwargs.get('unit_slug')

    if context['view'].kwargs.get('coa_slug'):
        kwargs['coa_slug'] = context['view'].kwargs.get('coa_slug')

    ctx = dict()
    ctx['year'] = context['year']
    ctx['has_year'] = context.get('has_year')
    ctx['has_quarter'] = context.get('has_quarter')
    ctx['has_month'] = context.get('has_month')
    ctx['has_date'] = context.get('has_date')
    ctx['previous_year'] = context['previous_year']

    kwargs['year'] = context['previous_year']
    ctx['previous_year_url'] = reverse(f'django_ledger:{base_url}-year', kwargs=kwargs)
    ctx['next_year'] = context['next_year']

    kwargs['year'] = context['next_year']
    ctx['next_year_url'] = reverse(f'django_ledger:{base_url}-year', kwargs=kwargs)

    kwargs['year'] = context['year']
    ctx['current_year_url'] = reverse(f'django_ledger:{base_url}-year', kwargs=kwargs)

    dt = get_localdate()

    KWARGS_CURRENT_MONTH = {
        'entity_slug': context['view'].kwargs['entity_slug'],
        'year': dt.year,
        'month': dt.month
    }

    if 'unit_slug' in kwargs:
        KWARGS_CURRENT_MONTH['unit_slug'] = kwargs['unit_slug']
    if 'account_pk' in kwargs:
        KWARGS_CURRENT_MONTH['account_pk'] = kwargs['account_pk']
    if 'ledger_pk' in kwargs:
        KWARGS_CURRENT_MONTH['ledger_pk'] = kwargs['ledger_pk']
    if 'coa_slug' in kwargs:
        KWARGS_CURRENT_MONTH['coa_slug'] = kwargs['coa_slug']

    ctx['current_month_url'] = reverse(f'django_ledger:{base_url}-month',
                                       kwargs=KWARGS_CURRENT_MONTH)

    quarter_urls = list()
    ctx['quarter'] = context.get('quarter')
    for Q in range(1, 5):
        kwargs['quarter'] = Q
        quarter_urls.append({
            'url': reverse(f'django_ledger:{base_url}-quarter', kwargs=kwargs),
            'quarter': Q,
            'quarter_name': f'Q{Q}'
        })
    del kwargs['quarter']
    ctx['quarter_urls'] = quarter_urls

    month_urls = list()
    ctx['month'] = context.get('month')
    for M in range(1, 13):
        kwargs['month'] = M
        month_urls.append({
            'url': reverse(f'django_ledger:{base_url}-month', kwargs=kwargs),
            'month': M,
            'month_abbr': month_abbr[M]
        })
    ctx['month_urls'] = month_urls
    ctx['from_date'] = context['from_date']
    ctx['to_date'] = context['to_date']
    ctx.update(kwargs)

    ctx['date_navigation_url'] = context.get('date_navigation_url')

    return ctx


@register.inclusion_tag('django_ledger/components/menu.html', takes_context=True)
def navigation_menu(context, style):
    ENTITY_SLUG = context['view'].kwargs.get('entity_slug')

    ctx = dict()
    ctx['style'] = style
    if ENTITY_SLUG:
        ctx['entity_slug'] = ENTITY_SLUG
        nav_menu_links = [
            {
                'type': 'link',
                'title': 'Entity Dashboard',
                'url': reverse('django_ledger:entity-dashboard', kwargs={'entity_slug': ENTITY_SLUG}),
                'icon': 'mdi:view-dashboard'
            },
            {
                'type': 'links',
                'title': 'Management',
                'links': [
                    {
                        'type': 'link',
                        'title': 'Vendors',
                        'url': reverse('django_ledger:vendor-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:truck'
                    },
                    {
                        'type': 'link',
                        'title': 'Customers',
                        'url': reverse('django_ledger:customer-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:account-group'
                    },
                    {
                        'type': 'link',
                        'title': 'Bank Accounts',
                        'url': reverse('django_ledger:bank-account-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:bank'
                    },
                    {
                        'type': 'link',
                        'title': 'Estimates & Contracts',
                        'url': reverse('django_ledger:customer-estimate-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:clipboard-text'
                    },
                    {
                        'type': 'link',
                        'title': 'Bills',
                        'url': reverse('django_ledger:bill-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'uil:bill'
                    },
                    {
                        'type': 'link',
                        'title': 'Invoices',
                        'url': reverse('django_ledger:invoice-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:file-document'
                    },
                    {
                        'type': 'link',
                        'title': 'Purchase Orders',
                        'url': reverse('django_ledger:po-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:cart'
                    },
                    {
                        'type': 'link',
                        'title': 'Inventory',
                        'url': reverse('django_ledger:inventory-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:warehouse'
                    },
                    {
                        'type': 'link',
                        'title': 'Closing Entries',
                        'url': reverse('django_ledger:closing-entry-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:book-lock'
                    }
                ]
            },
            {
                'type': 'links',
                'title': 'Your Lists',
                'links': [
                    {
                        'type': 'link',
                        'title': 'Entity Units',
                        'url': reverse('django_ledger:unit-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:domain'
                    },
                    {
                        'type': 'link',
                        'title': 'Products',
                        'url': reverse('django_ledger:product-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:package-variant'
                    },
                    {
                        'type': 'link',
                        'title': 'Services',
                        'url': reverse('django_ledger:service-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:briefcase'
                    },
                    {
                        'type': 'link',
                        'title': 'Business Expenses',
                        'url': reverse('django_ledger:expense-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:cash-minus'
                    },
                    {
                        'type': 'link',
                        'title': 'Inventory Items',
                        'url': reverse('django_ledger:inventory-item-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:barcode'
                    },
                    {
                        'type': 'link',
                        'title': 'Unit of Measures',
                        'url': reverse('django_ledger:uom-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:ruler'
                    },
                ]
            },
            {
                'type': 'links',
                'title': 'Reports',
                'links': [
                    {
                        'type': 'link',
                        'title': 'Balance Sheet',
                        'url': reverse('django_ledger:entity-bs', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'clarity:balance-line'
                    },
                    {
                        'type': 'link',
                        'title': 'Income Statement',
                        'url': reverse('django_ledger:entity-ic', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:chart-line'
                    },
                    {
                        'type': 'link',
                        'title': 'Cash Flow Statement',
                        'url': reverse('django_ledger:entity-cf', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'hugeicons:money-send-flow-02'
                    },
                ]
            },
            {
                'type': 'links',
                'title': 'Accounting',
                'links': [
                    {
                        'type': 'link',
                        'title': 'Chart of Accounts',
                        'url': reverse('django_ledger:coa-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'system-uicons:hierarchy'
                    },
                    {
                        'type': 'link',
                        'title': 'Ledgers',
                        'url': reverse('django_ledger:ledger-list-visible', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:book-open'
                    },
                    {
                        'type': 'link',
                        'title': 'Data Import',
                        'url': reverse('django_ledger:data-import-jobs-list', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:database-import'
                    }
                ]
            },
            {
                'type': 'links',
                'title': 'Administration',
                'links': [
                    {
                        'type': 'link',
                        'title': 'My Entities',
                        'url': reverse('django_ledger:home'),
                        'icon': 'mdi:home-group'
                    },
                    {
                        'type': 'link',
                        'title': 'Entity Settings',
                        'url': reverse('django_ledger:entity-update', kwargs={'entity_slug': ENTITY_SLUG}),
                        'icon': 'mdi:cog'
                    }
                ]
            }
        ]
        ctx['links'] = nav_menu_links
        ctx['request'] = context['request']
    return ctx

@register.inclusion_tag('django_ledger/product/tags/product_table.html', takes_context=True)
def product_table(context, queryset):
    entity_slug = context['view'].kwargs['entity_slug']
    return {
        'entity_slug': entity_slug,
        'product_list': queryset
    }


@register.inclusion_tag('django_ledger/service/tags/services_table.html', takes_context=True)
def service_table(context, queryset):
    entity_slug = context['view'].kwargs['entity_slug']
    return {
        'entity_slug': entity_slug,
        'service_list': queryset
    }


@register.inclusion_tag('django_ledger/expense/tags/expense_item_table.html', takes_context=True)
def expense_item_table(context, queryset):
    entity_slug = context['view'].kwargs['entity_slug']
    return {
        'entity_slug': entity_slug,
        'expense_list': queryset
    }


@register.inclusion_tag('django_ledger/inventory/tags/inventory_item_table.html', takes_context=True)
def inventory_item_table(context, queryset):
    entity_slug = context['view'].kwargs['entity_slug']
    return {
        'entity_slug': entity_slug,
        'inventory_item_list': queryset
    }


@register.inclusion_tag('django_ledger/invoice/tags/invoice_item_formset.html', takes_context=True)
def invoice_item_formset_table(context, itemtxs_formset):
    return {
        'entity_slug': context['view'].kwargs['entity_slug'],
        'invoice_model': context['invoice'],
        'total_amount__sum': context['total_amount__sum'],
        'itemtxs_formset': itemtxs_formset,
    }


@register.inclusion_tag('django_ledger/bills/tags/bill_item_formset.html', takes_context=True)
def bill_item_formset_table(context, item_formset):
    return {
        'entity_slug': context['view'].kwargs['entity_slug'],
        'bill_pk': context['view'].kwargs['bill_pk'],
        'total_amount__sum': context['total_amount__sum'],
        'item_formset': item_formset,
    }


@register.inclusion_tag('django_ledger/purchase_order/includes/po_item_formset.html', takes_context=True)
def po_item_formset_table(context, po_model, itemtxs_formset):
    return {
        'entity_slug': context['view'].kwargs['entity_slug'],
        'po_model': po_model,
        'itemtxs_formset': itemtxs_formset,
    }


@register.inclusion_tag('django_ledger/uom/tags/uom_table.html', takes_context=True)
def uom_table(context, uom_queryset):
    return {
        'entity_slug': context['view'].kwargs['entity_slug'],
        'uom_list': uom_queryset
    }


@register.inclusion_tag('django_ledger/inventory/tags/inventory_table.html', takes_context=True)
def inventory_table(context, queryset):
    ctx = {
        'entity_slug': context['view'].kwargs['entity_slug'],
        'inventory_list': queryset
    }
    ctx.update(queryset.aggregate(inventory_total_value=Sum('total_value')))
    return ctx


@register.inclusion_tag('django_ledger/estimate/includes/estimate_table.html', takes_context=True)
def customer_estimate_table(context, queryset):
    return {
        'entity_slug': context['view'].kwargs['entity_slug'],
        'ce_list': queryset
    }


@register.inclusion_tag('django_ledger/estimate/includes/estimate_item_table.html', takes_context=True)
def estimate_item_table(context, queryset):
    return {
        'entity_slug': context['view'].kwargs['entity_slug'],
        'ce_model': context['estimate_model'],
        'ce_item_list': queryset
    }


@register.inclusion_tag('django_ledger/purchase_order/tags/po_item_table.html', takes_context=True)
def po_item_table(context, queryset):
    return {
        'entity_slug': context['view'].kwargs['entity_slug'],
        'po_model': context['po_model'],
        'po_item_list': queryset
    }


@register.inclusion_tag('django_ledger/estimate/tags/ce_item_formset.html', takes_context=True)
def customer_estimate_item_formset(context, item_formset):
    return {
        'entity_slug': context['view'].kwargs['entity_slug'],
        'ce_pk': context['view'].kwargs['ce_pk'],
        'ce_revenue_estimate__sum': context['ce_revenue_estimate__sum'],
        'ce_cost_estimate__sum': context['ce_cost_estimate__sum'],
        'item_formset': item_formset,
    }



from django.forms import DateInput, ValidationError, ModelForm, Textarea, HiddenInput
from django.utils.translation import gettext_lazy as _

from django_ledger.io.io_core import get_localdate
from django_ledger.models.closing_entry import ClosingEntryModel
from django_ledger.models.entity import EntityModel
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class ClosingEntryCreateForm(ModelForm):

    def __init__(self, entity_model: EntityModel, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ENTITY_MODEL: EntityModel = entity_model
        self.fields['entity_model'].disabled = True

    def clean_entity_model(self):
        return self.ENTITY_MODEL

    def clean_closing_date(self):
        closing_date = self.cleaned_data['closing_date']

        if closing_date > get_localdate():
            raise ValidationError(
                message=_('Cannot create a closing entry with a future date.'), code='invalid_date'
            )
        return closing_date

    class Meta:
        model = ClosingEntryModel
        fields = [
            'entity_model',
            'closing_date'
        ]

        widgets = {
            'entity_model': HiddenInput(),
            'closing_date': DateInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-large',
                'placeholder': _('Closing Date (YYYY-MM-DD)...'),
                'id': 'djl-datepicker',
                'type': 'date'
            })
        }
        labels = {
            'closing_date': _('Select a Closing Date')
        }


class ClosingEntryUpdateForm(ModelForm):
    class Meta:
        model = ClosingEntryModel
        fields = [
            'markdown_notes'
        ]

        widgets = {
            'markdown_notes': Textarea(attrs={
                'class': 'textarea'
            }),
        }
        labels = {
            'markdown_notes': _('Closing Entry Notes')
        }



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Chart Of Accounts
-----------------

A Chart of Accounts (CoA) is a fundamental component of financial management in Django Ledger. It serves as the
backbone of financial statements and is organized within a ChartOfAccountModel.

### Key Features

- **Account Roles**: The CoA includes various account types such as:
  - Cash
  - Accounts Receivable
  - Expenses
  - Liabilities
  - Income

- **Hierarchical Structure**: Accounts are logically grouped to form financial statements. For example, the Balance
Sheet may have a structure like this:
  - Fixed Assets
    - Tangible Assets
      - Building
      - Plant & Equipment
      - Machinery
    - Intangible Assets

- **Financial Statement Preparation**: Individual account balances are aggregated based on the CoA and AccountModel
roles to create comprehensive financial statements.

### Usage in EntityModel

- Every EntityModel must have a default CoA to create any type of transaction.
- If no explicit CoA is specified, the EntityModel's default CoA is used.
- Only ONE Chart of Accounts can be used when creating Journal Entries.
- Commingling between different CoAs is not allowed to maintain the integrity of Journal Entries.

This structure ensures a clear and organized approach to financial management within Django Ledger, facilitating
accurate record-keeping and reporting.
"""
import warnings
from random import choices
from string import ascii_lowercase, digits
from typing import Optional, Union, Dict
from uuid import uuid4, UUID

from django.apps import apps
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q, F, Count, Manager, QuerySet, BooleanField, Value
from django.db.models.signals import pre_save, post_save
from django.dispatch import receiver
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from django_ledger.io import (
    ROOT_COA, ROOT_GROUP_LEVEL_2, ROOT_GROUP_META, ROOT_ASSETS,
    ROOT_LIABILITIES, ROOT_CAPITAL,
    ROOT_INCOME, ROOT_COGS, ROOT_EXPENSES, ROOT_GROUP
)
from django_ledger.models import lazy_loader
from django_ledger.models.accounts import AccountModel, AccountModelQuerySet
from django_ledger.models.deprecations import deprecated_entity_slug_behavior
from django_ledger.models.mixins import CreateUpdateMixIn, SlugNameMixIn
from django_ledger.settings import DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR

UserModel = get_user_model()

SLUG_SUFFIX = ascii_lowercase + digits

app_config = apps.get_app_config('django_ledger')


class ChartOfAccountsModelValidationError(ValidationError):
    pass


class ChartOfAccountModelQuerySet(QuerySet):

    def active(self) -> 'ChartOfAccountModelQuerySet':
        """
        QuerySet method to retrieve active items.
        """
        return self.filter(active=True)

    def not_active(self) -> 'ChartOfAccountModelQuerySet':
        """≤
        QuerySet method to retrieve not active items.
        """
        return self.filter(active=False)

    def for_user(self, user_model) -> 'ChartOfAccountModelQuerySet':
        """
        Fetches a QuerySet of ChartOfAccountModel that the UserModel as access to. May include ChartOfAccountModel from
        multiple Entities. The user has access to bills if:
        1. Is listed as Manager of Entity.
        2. Is the Admin of the Entity.

        Parameters
        ----------
        user_model
            Logged in and authenticated django UserModel instance.

        Returns
        -------
        ChartOfAccountQuerySet
            Returns a ChartOfAccountQuerySet with applied filters.
        """

        if user_model.is_superuser:
            return self

        return self.filter(
            (
                    Q(entity__admin=user_model) |
                    Q(entity__managers__in=[user_model])
            )
        )


class ChartOfAccountModelManager(Manager):
    """
    A custom-defined ChartOfAccountModelManager that will act as an interface to handling the initial DB queries
    to the ChartOfAccountModel.
    """

    def get_queryset(self) -> ChartOfAccountModelQuerySet:
        qs = ChartOfAccountModelQuerySet(self.model, using=self._db)
        return qs.annotate(
            _entity_slug=F('entity__slug'),
            accountmodel_total__count=Count(
                'accountmodel',
                # excludes coa root accounts...
                filter=Q(accountmodel__depth__gt=2)
            ),
            accountmodel_locked__count=Count(
                'accountmodel',
                # excludes coa root accounts...
                filter=Q(accountmodel__depth__gt=2) & Q(accountmodel__locked=True)
            ),
            accountmodel_active__count=Count(
                'accountmodel',
                # excludes coa root accounts...
                filter=Q(accountmodel__depth__gt=2) & Q(accountmodel__active=True)
            ),
            # Root-group presence and uniqueness checks:
            accountmodel_rootgroup__count=Count(
                'accountmodel',
                filter=Q(accountmodel__role__in=ROOT_GROUP)
            ),
            accountmodel_rootgroup_roles__distinct_count=Count(
                'accountmodel__role',
                filter=Q(accountmodel__role__in=ROOT_GROUP_META),
                distinct=True
            ),
        ).annotate(
            configured=models.Case(
                models.When(
                    Q(accountmodel_rootgroup__count__gte=1) &
                    Q(accountmodel_rootgroup__count=F('accountmodel_rootgroup_roles__distinct_count')),
                    then=Value(True, output_field=BooleanField()),
                ),
                default=Value(False, output_field=BooleanField()),
                output_field=BooleanField()
            )
        ).select_related('entity')

    @deprecated_entity_slug_behavior
    def for_entity(self,
                   entity_model: Union['EntityModel | str | UUID'] = None,
                   **kwargs) -> ChartOfAccountModelQuerySet:
        """
        Fetches a QuerySet of ChartOfAccountsModel associated with a specific EntityModel & UserModel.
        May pass an instance of EntityModel or a String representing the EntityModel slug.

        Parameters
        __________

        entity_slug: str or EntityModel
            The entity slug or EntityModel used for filtering the QuerySet.

        Returns
        -------
        ChartOfAccountQuerySet
            Returns a ChartOfAccountQuerySet with applied filters.
        """

        EntityModel = lazy_loader.get_entity_model()

        qs = self.get_queryset()
        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(entity=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(entity__slug=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(entity_id=entity_model)
        else:
            raise ChartOfAccountsModelValidationError(
                message='Must pass an instance of EntityModel, String or UUID for entity_slug.'
            )
        return qs


class ChartOfAccountModelAbstract(SlugNameMixIn, CreateUpdateMixIn):
    """
    Abstract base class for the Chart of Account model.

    Attributes
    ----------
    uuid : UUIDField
        UUID field for the chart of account model (primary key).
    entity : ForeignKey
        ForeignKey to the EntityModel.
    active : BooleanField
        BooleanField indicating whether the chart of account is active or not.
    description : TextField
        TextField storing the description of the chart of account.
    objects : ChartOfAccountModelManager
        Manager for the ChartOfAccountModel.
    """

    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    entity = models.ForeignKey('django_ledger.EntityModel',
                               verbose_name=_('Entity'),
                               on_delete=models.CASCADE)
    active = models.BooleanField(default=True, verbose_name=_('Is Active'))
    description = models.TextField(verbose_name=_('CoA Description'), null=True, blank=True)
    objects = ChartOfAccountModelManager.from_queryset(queryset_class=ChartOfAccountModelQuerySet)()

    class Meta:
        abstract = True
        ordering = ['-created']
        verbose_name = _('Chart of Account')
        verbose_name_plural = _('Chart of Accounts')
        indexes = [
            models.Index(fields=['entity'])
        ]

    def __str__(self):
        if self.name is not None:
            return f'{self.name} ({self.slug})'
        return self.slug

    @property
    def entity_slug(self) -> str:
        try:
            return getattr(self, '_entity_slug')
        except AttributeError:
            return self.entity.slug

    def is_configured(self) -> bool:
        try:
            return getattr(self, 'configured')
        except AttributeError:
            pass
        account_qs = self.accountmodel_set.filter(role__in=[ROOT_GROUP])
        self.configured = len(account_qs) == len(ROOT_GROUP)
        return self.configured

    def configure(self, raise_exception: bool = True):
        """
        A method that properly configures the ChartOfAccounts model and creates the appropriate hierarchy boilerplate
        to support the insertion of new accounts into the chart of account model tree.
        This method must be called every time the ChartOfAccounts model is created.

        Parameters
        ----------
        raise_exception : bool, optional
            Whether to raise an exception if root nodes already exist in the Chart of Accounts (default is True).
            This indicates that the ChartOfAccountModel instance is already configured.
        """
        self.generate_slug(commit=False)

        if not self.is_configured():
            root_accounts_qs = self.get_coa_root_accounts_qs()
            existing_root_roles = list(set(acc.role for acc in root_accounts_qs))

            if len(existing_root_roles) > 0:
                if raise_exception:
                    raise ChartOfAccountsModelValidationError(message=f'Root Nodes already Exist in CoA {self.uuid}...')
                return

            if ROOT_COA not in existing_root_roles:
                # add coa root...
                role_meta = ROOT_GROUP_META[ROOT_COA]
                account_pk = uuid4()
                root_account = AccountModel(
                    uuid=account_pk,
                    code=role_meta['code'],
                    name=role_meta['title'],
                    coa_model=self,
                    role=ROOT_COA,
                    role_default=True,
                    active=False,
                    locked=True,
                    balance_type=role_meta['balance_type']
                )
                AccountModel.add_root(instance=root_account)

                # must retrieve root model after added pero django-treebeard documentation...
                coa_root_account_model = AccountModel.objects.get(uuid__exact=account_pk)

                for root_role in ROOT_GROUP_LEVEL_2:
                    if root_role not in existing_root_roles:
                        account_pk = uuid4()
                        role_meta = ROOT_GROUP_META[root_role]
                        coa_root_account_model.add_child(
                            instance=AccountModel(
                                uuid=account_pk,
                                code=role_meta['code'],
                                name=role_meta['title'],
                                coa_model=self,
                                role=root_role,
                                role_default=True,
                                active=False,
                                locked=True,
                                balance_type=role_meta['balance_type']
                            ))
                self.configured = True

    def get_coa_root_accounts_qs(self) -> AccountModelQuerySet:
        """
        Retrieves the root accounts in the chart of accounts.

        Returns:
            AccountModelQuerySet: A queryset containing the root accounts in the chart of accounts.
        """
        return self.accountmodel_set.all().is_coa_root()

    def get_coa_root_node(self) -> AccountModel:
        """
        Retrieves the root node of the chart of accounts.

        Returns:
            AccountModel: The root node of the chart of accounts.

        """
        qs = self.get_coa_root_accounts_qs()
        return qs.get(role__exact=ROOT_COA)

    def get_account_root_node(self,
                              account_model: AccountModel,
                              root_account_qs: Optional[AccountModelQuerySet] = None,
                              as_queryset: bool = False) -> AccountModel:
        """
        Fetches the root node of the ChartOfAccountModel instance. The root node is the highest level of the CoA
        hierarchy. It can be used to traverse the hierarchy of the CoA structure downstream.


        Parameters
        ----------
        account_model : AccountModel
            The account model for which to find the root node.
        root_account_qs : Optional[AccountModelQuerySet], optional
            The queryset of root accounts. If not provided, it will be retrieved using `get_coa_root_accounts_qs` method.
        as_queryset : bool, optional
            If True, return the root account queryset instead of a single root account. Default is False.

        Returns
        -------
        Union[AccountModelQuerySet, AccountModel]
            If `as_queryset` is True, returns the root account queryset. Otherwise, returns a single root account.

        Raises
        ------
        ChartOfAccountsModelValidationError
            If the account model is not part of the chart of accounts.
        """

        if account_model.coa_model_id != self.uuid:
            raise ChartOfAccountsModelValidationError(
                message=_(f'The account model {account_model} is not part of the chart of accounts {self.name}.'),
            )

        # Chart of Accounts hasn't been configured...
        if not self.is_configured():
            self.configure(raise_exception=True)

        if not account_model.is_root_account():

            if not root_account_qs:
                root_account_qs = self.get_coa_root_accounts_qs()

            if account_model.is_asset():
                qs = root_account_qs.filter(code__exact=ROOT_GROUP_META[ROOT_ASSETS]['code'])
            elif account_model.is_liability():
                qs = root_account_qs.filter(code__exact=ROOT_GROUP_META[ROOT_LIABILITIES]['code'])
            elif account_model.is_capital():
                qs = root_account_qs.filter(code__exact=ROOT_GROUP_META[ROOT_CAPITAL]['code'])
            elif account_model.is_income():
                qs = root_account_qs.filter(code__exact=ROOT_GROUP_META[ROOT_INCOME]['code'])
            elif account_model.is_cogs():
                qs = root_account_qs.filter(code__exact=ROOT_GROUP_META[ROOT_COGS]['code'])
            elif account_model.is_expense():
                qs = root_account_qs.filter(code__exact=ROOT_GROUP_META[ROOT_EXPENSES]['code'])
            else:
                raise ChartOfAccountsModelValidationError(message=f'Unable to locate Balance Sheet'
                                                                  ' root node for account code: '
                                                                  f'{account_model.code} {account_model.name}')
            if as_queryset:
                return qs
            return qs.get()

        raise ChartOfAccountsModelValidationError(
            message='Adding Root account to Chart of Accounts is not allowed.'
        )

    def get_non_root_coa_accounts_qs(self) -> AccountModelQuerySet:
        """
        Returns a query set of non-root accounts in the chart of accounts.

        Returns
        -------
        AccountModelQuerySet
            A query set of non-root accounts in the chart of accounts.
        """
        return self.accountmodel_set.all().not_coa_root()

    def get_coa_accounts(self, active_only: bool = True) -> AccountModelQuerySet:
        """
        Returns the AccountModelQuerySet associated with the ChartOfAccounts model instance.

        Parameters
        ----------
        active_only : bool, optional
            Flag to indicate whether to retrieve only active accounts or all accounts.
            Default is True.

        Returns
        -------
        AccountModelQuerySet
            A queryset containing accounts from the chart of accounts.

        """
        qs = self.get_non_root_coa_accounts_qs()
        if active_only:
            return qs.active()
        return qs

    def get_coa_account_tree(self) -> Dict:
        """
        Performs a bulk dump of the ChartOfAccounts model instance accounts to a dictionary.
        The method invokes the`dump_bulk` method on the ChartOfAccount model instance root node.
        See Django Tree Beard documentation for more information.

        Returns
        -------
        Dict
            A dictionary containing all accounts from the chart of accounts in a nested structure.
        """
        root_account = self.get_coa_root_node()
        return AccountModel.dump_bulk(parent=root_account)

    def generate_slug(self, commit: bool = False, raise_exception: bool = False) -> str:
        """
        Generates and assigns a slug based on the ChartOfAccounts model instance EntityModel information.


        Parameters
        ----------
        raise_exception : bool, optional
                If set to True, it will raise a ChartOfAccountsModelValidationError if the `self.slug` is already set.

        Returns
        -------
        str
                The generated slug for the Chart of Accounts.

        Raises
        ------
        ChartOfAccountsModelValidationError
                If `raise_exception` is set to True and `self.slug` is already set.

        """
        if self.slug:
            if raise_exception:
                raise ChartOfAccountsModelValidationError(
                    message=_(f'CoA {self.uuid} already has a slug')
                )
            return
        self.slug = f'coa-{self.entity.slug[-5:]}-' + ''.join(choices(SLUG_SUFFIX, k=15))

        if commit:
            self.save(
                update_fields=[
                    'slug',
                    'updated'
                ]
            )

    def is_default(self) -> bool:
        """
        Check if the ChartOfAccountModel instance is set as the default for the EntityModel.

        Returns
        -------
        bool
            True if the ChartOfAccountModel instance is set as the default for the EntityModel. Else, False.
        """
        if not self.entity_id:
            return False
        if not self.entity.default_coa_id:
            return False
        return self.entity.default_coa_id == self.uuid

    def is_active(self) -> bool:
        """
        Check if the ChartOfAccountModel instance is active.

        Returns:
            bool: True if the ChartOfAccountModel instance is active, False otherwise.
        """
        return self.active is True

    def validate_account_model_qs(self, account_model_qs: AccountModelQuerySet):
        """
        Validates the given AccountModelQuerySet for the ChartOfAccountsModel.

        Parameters
        ----------
        account_model_qs : AccountModelQuerySet
            The AccountModelQuerySet to validate.

        Raises
        ------
        ChartOfAccountsModelValidationError
            If the account_model_qs is not an instance of AccountModelQuerySet or if it contains an account model with a different coa_model_id than the current CoA model.

        """
        if not isinstance(account_model_qs, AccountModelQuerySet):
            raise ChartOfAccountsModelValidationError(
                message='Must pass an instance of AccountModelQuerySet'
            )
        for acc_model in account_model_qs:
            if not acc_model.coa_model_id == self.uuid:
                raise ChartOfAccountsModelValidationError(
                    message=f'Invalid root queryset for CoA {self.name}'
                )

    def insert_account(self,
                       account_model: AccountModel,
                       root_account_qs: Optional[AccountModelQuerySet] = None):
        """
        This method inserts the given account model into the chart of accounts (COA) instance.
        It first verifies if the account model's COA model ID matches the COA's UUID. If not, it
        raises a `ChartOfAccountsModelValidationError`. If the `root_account_qs` is not provided, it retrieves the
        root account query set using the `get_coa_root_accounts_qs` method. Providing a pre-fetched `root_account_qs`
        avoids unnecessary retrieval of the root account query set every an account model is inserted into the CoA.

        Next, it validates the provided `root_account_qs` if it is not None. Then, it obtains the root node for the
        account model using the `get_account_root_node` method and assigns it to `account_root_node`.

        Finally, it adds the account model as a child to the `account_root_node` and retrieves the updated COA accounts
        query set using the `get_non_root_coa_accounts_qs` method. It returns the inserted account model found in the
        COA accounts query set.

        Parameters
        ----------
        account_model : AccountModel
            The account model to be inserted into the chart of accounts.
        root_account_qs : Optional[AccountModelQuerySet], default=None
            The root account query set. If not provided, it will be obtained using the `get_coa_root_accounts_qs`
            method.

        Returns
        -------
        AccountModel
            The inserted account model.

        Raises
        ------
        ChartOfAccountsModelValidationError
            If the provided account model has an invalid COA model ID for the current COA.
        """

        if account_model.coa_model_id:
            if account_model.coa_model_id != self.uuid:
                raise ChartOfAccountsModelValidationError(
                    message=f'Invalid Account Model {account_model} for CoA {self}'
                )
        else:
            account_model.coa_model = self

        if not root_account_qs:
            root_account_qs = self.get_coa_root_accounts_qs()
        else:
            self.validate_account_model_qs(root_account_qs)

        account_root_node: AccountModel = self.get_account_root_node(
            account_model=account_model,
            root_account_qs=root_account_qs
        )

        account_root_node.add_child(instance=account_model)
        coa_accounts_qs = self.get_non_root_coa_accounts_qs()
        return coa_accounts_qs.get(uuid__exact=account_model.uuid)

    def create_account(self,
                       code: str,
                       role: str,
                       name: str,
                       balance_type: str,
                       active: bool,
                       root_account_qs: Optional[AccountModelQuerySet] = None):
        """
        Proper method for inserting a new Account Model into a CoA.
        Use this in liu of the direct instantiation of the AccountModel of using the django related manager.

        Parameters
        ----------
        code : str
            The code of the account to be created.
        role : str
            The role of the account. This can be a user-defined value.
        name : str
            The name of the account.
        balance_type : str
            The balance type of the account. This can be a user-defined value.
        active : bool
            Specifies whether the account is active or not.
        root_account_qs : Optional[AccountModelQuerySet], optional
            The query set of root accounts to which the created account should be linked. Defaults to None.

        Returns
        -------
        AccountModel
            The created account model instance.
        """
        account_model = AccountModel(
            code=code,
            name=name,
            role=role,
            active=active,
            balance_type=balance_type,
            coa_model=self
        )
        account_model.clean()

        account_model = self.insert_account(
            account_model=account_model,
            root_account_qs=root_account_qs
        )
        return account_model

    # ACTIONS -----
    # todo: use these methods once multi CoA features are enabled...
    def lock_all_accounts(self) -> AccountModelQuerySet:
        account_qs = self.get_coa_accounts()
        account_qs.update(locked=True)
        return account_qs

    def unlock_all_accounts(self) -> AccountModelQuerySet:
        account_qs = self.get_non_root_coa_accounts_qs()
        account_qs.update(locked=False)
        return account_qs

    def mark_as_default(self, commit: bool = False, raise_exception: bool = False, **kwargs):
        """
        Marks the current Chart of Accounts instances as default for the EntityModel.

        Parameters
        ----------
        commit: bool
            Commit the action into the Database. Default is False.
        raise_exception: bool
            Raises exception if Chart of Account model instance is already marked as default.
        """
        if self.is_default():
            if raise_exception:
                raise ChartOfAccountsModelValidationError(
                    message=_(f'The Chart of Accounts {self.slug} is already default')
                )
            return
        if not self.can_mark_as_default():
            if raise_exception:
                raise ChartOfAccountsModelValidationError(
                    message=_(f'The Chart of Accounts {self.slug} cannot be marked as default')
                )
            return
        self.entity.default_coa_id = self.uuid
        self.clean()
        if commit:
            self.entity.save(
                update_fields=[
                    'default_coa_id',
                    'updated'
                ]
            )

    def can_mark_as_default(self):
        return all([
            self.is_active(),
            not self.is_default()
        ])

    def can_activate(self) -> bool:
        """
        Check if the ChartOffAccountModel instance can be activated.

        Returns
        -------
            True if the object can be activated, False otherwise.
        """
        return self.active is False

    def can_deactivate(self) -> bool:
        """
        Check if the ChartOffAccountModel instance can be deactivated.

        Returns
        -------
            True if the object can be deactivated, False otherwise.
        """
        return all([
            self.is_active(),
            not self.is_default()
        ])

    def mark_as_active(self, commit: bool = False, raise_exception: bool = False, **kwargs):
        """
        Marks the current Chart of Accounts as Active.

        Parameters
        ----------
        commit: bool
            Commit the action into the Database. Default is False.
        raise_exception: bool
            Raises exception if Chart of Account model instance is already active. Default is False.
        """
        if self.is_active():
            if raise_exception:
                raise ChartOfAccountsModelValidationError(
                    message=_('The Chart of Accounts is currently active.')
                )
            return

        self.active = True
        self.clean()
        if commit:
            self.save(
                update_fields=[
                    'active',
                    'updated'
                ])

    def mark_as_inactive(self, commit: bool = False, raise_exception: bool = False, **kwargs):
        """
        Marks the current Chart of Accounts as Active.

        Parameters
        ----------
        commit: bool
            Commit the action into the Database. Default is False.
        raise_exception: bool
            Raises exception if Chart of Account model instance is already active. Default is False.
        """
        if not self.is_active():
            if raise_exception:
                raise ChartOfAccountsModelValidationError(
                    message=_('The Chart of Accounts is currently not active.')
                )
            return

        self.active = False
        self.clean()
        if commit:
            self.save(
                update_fields=[
                    'active',
                    'updated'
                ])

    # URLS....
    def mark_as_default_url(self) -> str:
        """
        Returns the URL to mark the current Chart of Accounts instances as Default for the EntityModel.

        Returns
        -------
        str
            The URL as a String.
        """
        return reverse(
            viewname='django_ledger:coa-action-mark-as-default',
            kwargs={
                'entity_slug': self.entity_slug,
                'coa_slug': self.slug
            }
        )

    def mark_as_active_url(self) -> str:
        """
        Returns the URL to mark the current Chart of Accounts instances as active.

        Returns
        -------
        str
            The URL as a String.
        """
        return reverse(
            viewname='django_ledger:coa-action-mark-as-active',
            kwargs={
                'entity_slug': self.entity_slug,
                'coa_slug': self.slug
            }
        )

    def mark_as_inactive_url(self) -> str:
        """
        Returns the URL to mark the current Chart of Accounts instances as inactive.

        Returns
        -------
        str
            The URL as a String.
        """
        return reverse(
            viewname='django_ledger:coa-action-mark-as-inactive',
            kwargs={
                'entity_slug': self.entity_slug,
                'coa_slug': self.slug
            }
        )

    def get_coa_list_url(self):
        return reverse(
            viewname='django_ledger:coa-list',
            kwargs={
                'entity_slug': self.entity_slug
            }
        )

    def get_coa_list_inactive_url(self):
        return reverse(
            viewname='django_ledger:coa-list-inactive',
            kwargs={
                'entity_slug': self.entity_slug
            }
        )

    def get_coa_create_url(self):
        return reverse(
            viewname='django_ledger:coa-create',
            kwargs={
                'entity_slug': self.entity_slug
            }
        )

    def get_absolute_url(self) -> str:
        return reverse(
            viewname='django_ledger:coa-detail',
            kwargs={
                'entity_slug': self.entity_slug,
                'coa_slug': self.slug
            }
        )

    def get_update_url(self) -> str:
        return reverse(
            viewname='django_ledger:coa-update',
            kwargs={
                'entity_slug': self.entity_slug,
                'coa_slug': self.slug
            }
        )

    def get_account_list_url(self):

        if not self.slug:
            self.generate_slug(commit=True)

        return reverse(
            viewname='django_ledger:account-list',
            kwargs={
                'entity_slug': self.entity_slug,
                'coa_slug': self.slug
            }
        )

    def get_create_coa_account_url(self):
        return reverse(
            viewname='django_ledger:account-create',
            kwargs={
                'coa_slug': self.slug,
                'entity_slug': self.entity_slug
            }
        )

    def clean(self):
        self.generate_slug()


class ChartOfAccountModel(ChartOfAccountModelAbstract):
    """
    Base ChartOfAccounts Model
    """

    class Meta(ChartOfAccountModelAbstract.Meta):
        abstract = False


@receiver(pre_save, sender=ChartOfAccountModel)
def chartofaccountsmodel_presave(instance: ChartOfAccountModelAbstract, **kwargs):
    instance.generate_slug()
    if instance.is_default() and not instance.active:
        raise ChartOfAccountsModelValidationError(
            _('Default Chart of Accounts cannot be deactivated.')
        )


@receiver(post_save, sender=ChartOfAccountModel)
def chartofaccountsmodel_postsave(instance: ChartOfAccountModelAbstract, **kwargs):
    if not instance.is_configured():
        instance.configure()



# Generated by Django 5.2.1 on 2025-06-24 21:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0021_alter_bankaccountmodel_account_model_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='bankaccountmodel',
            name='financial_institution',
            field=models.CharField(blank=True, help_text='Name of the financial institution (i.e. Bank Name).', max_length=100, null=True, verbose_name='Financial Institution'),
        ),
        migrations.AddField(
            model_name='vendormodel',
            name='financial_institution',
            field=models.CharField(blank=True, help_text='Name of the financial institution (i.e. Bank Name).', max_length=100, null=True, verbose_name='Financial Institution'),
        ),
        migrations.AlterField(
            model_name='bankaccountmodel',
            name='account_type',
            field=models.CharField(choices=[('checking', 'Checking'), ('savings', 'Savings'), ('money_market', 'Money Market'), ('cert_deposit', 'Certificate of Deposit'), ('credit_card', 'Credit Card'), ('st_loan', 'Short Term Loan'), ('lt_loan', 'Long Term Loan'), ('mortgage', 'Mortgage'), ('other', 'Other')], default='checking', max_length=20, verbose_name='Account Type'),
        ),
        migrations.AlterField(
            model_name='billmodel',
            name='bill_items',
            field=models.ManyToManyField(through='django_ledger.ItemTransactionModel', through_fields=('bill_model', 'item_model'), to='django_ledger.itemmodel', verbose_name='Bill Items'),
        ),
        migrations.AlterField(
            model_name='invoicemodel',
            name='invoice_items',
            field=models.ManyToManyField(through='django_ledger.ItemTransactionModel', through_fields=('invoice_model', 'item_model'), to='django_ledger.itemmodel', verbose_name='Invoice Items'),
        ),
        migrations.AlterField(
            model_name='purchaseordermodel',
            name='po_items',
            field=models.ManyToManyField(through='django_ledger.ItemTransactionModel', through_fields=('po_model', 'item_model'), to='django_ledger.itemmodel', verbose_name='Purchase Order Items'),
        ),
        migrations.AlterField(
            model_name='vendormodel',
            name='account_type',
            field=models.CharField(choices=[('checking', 'Checking'), ('savings', 'Savings'), ('money_market', 'Money Market'), ('cert_deposit', 'Certificate of Deposit'), ('credit_card', 'Credit Card'), ('st_loan', 'Short Term Loan'), ('lt_loan', 'Long Term Loan'), ('mortgage', 'Mortgage'), ('other', 'Other')], default='checking', max_length=20, verbose_name='Account Type'),
        ),
        migrations.AddIndex(
            model_name='bankaccountmodel',
            index=models.Index(fields=['entity_model'], name='django_ledg_entity__6ad006_idx'),
        ),
    ]



from django.forms import Form, ModelChoiceField, Select, ChoiceField, CharField, HiddenInput, DateField, DateInput
from django.utils.translation import gettext_lazy as _l

from django_ledger.models.entity import EntityModel
from django_ledger.models.journal_entry import JournalEntryModel
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class EntityFilterForm(Form):
    entity_model = ModelChoiceField(
        queryset=EntityModel.objects.none(),
        widget=Select(attrs={
            'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' djetler-set-entity-form-input',
        }))

    def __init__(self, *args, user_model, current_entity_uuid=None, form_id=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.USER_MODEL = user_model
        self.form_id = form_id
        self.fields['entity_model'].queryset = EntityModel.objects.for_user(
            user_model=self.USER_MODEL).only('slug', 'name')
        if form_id:
            self.fields['entity_model'].widget.attrs['class'] += f' djetler-default-entity-input-{self.form_id}'

        if current_entity_uuid:
            self.initial = {
                'entity_model': current_entity_uuid
            }


class ActivityFilterForm(Form):
    CHOICES = [('all', _l('All'))] + JournalEntryModel.ACTIVITIES
    activity = ChoiceField(choices=CHOICES,
                           label=_l('Activity'),
                           initial='all',
                           widget=Select(
                               attrs={
                                   'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small djetler-activity-select-form-input',
                               }
                           ))


# class AsOfDateFilterForm(Form):
#
#     def __init__(self, *args, form_id=None, **kwargs):
#         super().__init__(*args, **kwargs)
#         self.form_id = form_id
#         if form_id:
#             self.fields['date'].widget.attrs['class'] += f' djetler-end-date-filter-input-{self.form_id}'
#
#     entity_slug = CharField(
#         max_length=150,
#         widget=HiddenInput()
#     )
#     date = DateField(
#         widget=DateInput(
#             attrs={
#                 'class': 'is-hidden',
#                 'data-input': True,
#             }
#         ))



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""



# Generated by Django 5.1.6 on 2025-02-18 22:47

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('django_ledger', '0020_remove_bankaccountmodel_django_ledg_cash_ac_59a8af_idx_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='bankaccountmodel',
            name='account_model',
            field=models.ForeignKey(
                help_text='Account model be used to map transactions from financial institution',
                on_delete=django.db.models.deletion.RESTRICT,
                to='django_ledger.AccountModel',
                verbose_name='Associated Account Model'
            ),
        ),
        migrations.AlterField(
            model_name='importjobmodel',
            name='bank_account_model',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to='django_ledger.BankAccountModel',
                verbose_name='Associated Bank Account Model'
            ),
        ),
    ]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

AccountModel
------------
The AccountModel is a fundamental component of the Django Ledger system, responsible for categorizing and organizing
financial transactions related to an entity's assets, liabilities, and equity.

Account Types
-------------

In accordance with accounting principles, each AccountModel must be classified as either:

1. **DEBIT-type balance account**
2. **CREDIT-type balance account**

The account type determines how transactions affect the account's balance.

Double Entry Accounting
-----------------------

The AccountModel is crucial in implementing double entry accounting systems:

* For DEBIT-type accounts:
  - A DEBIT increases the balance
  - A CREDIT decreases the balance

* For CREDIT-type accounts:
  - A CREDIT increases the balance
  - A DEBIT decreases the balance

Chart of Accounts
-----------------

Users have the flexibility to adopt a chart of accounts that best suits their EntityModel. Django Ledger provides a
default Chart of Accounts when creating a new EntityModel, which can be customized as needed.

Account Roles
-------------

All AccountModels must be assigned a role from the `ACCOUNT_ROLES` function in `django_ledger.io.roles`.
Roles serve several purposes:

1. Group accounts into common namespaces
2. Provide consistency across user-defined fields
3. Enable accurate generation of financial statements
4. Facilitate financial ratio calculations
"""
import warnings
from itertools import groupby
from random import randint
from typing import Union, List, Optional
from uuid import uuid4, UUID

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q, F, UniqueConstraint
from django.db.models.signals import pre_save
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from treebeard.mp_tree import MP_Node, MP_NodeManager, MP_NodeQuerySet

from django_ledger.io import DEBIT, CREDIT
from django_ledger.io.roles import (
    ACCOUNT_ROLE_CHOICES, BS_ROLES, GROUP_INVOICE, GROUP_BILL, validate_roles,
    GROUP_ASSETS, GROUP_LIABILITIES, GROUP_CAPITAL, GROUP_INCOME, GROUP_EXPENSES, GROUP_COGS,
    ROOT_GROUP, BS_BUCKETS, ROOT_ASSETS, ROOT_LIABILITIES,
    ROOT_CAPITAL, ROOT_INCOME, ROOT_EXPENSES, ROOT_COA, VALID_PARENTS,
    ROLES_ORDER_ALL, ASSET_CA_CASH
)
from django_ledger.models.deprecations import deprecated_entity_slug_behavior
from django_ledger.models.mixins import CreateUpdateMixIn
from django_ledger.models.utils import lazy_loader
from django_ledger.settings import (
    DJANGO_LEDGER_ACCOUNT_CODE_GENERATE,
    DJANGO_LEDGER_ACCOUNT_CODE_USE_PREFIX,
    DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR
)


class AccountModelValidationError(ValidationError):
    pass


class AccountModelQuerySet(MP_NodeQuerySet):
    """
    Custom QuerySet for AccountModel inheriting from MP_NodeQuerySet.
    """

    def active(self) -> 'AccountModelQuerySet':
        """
        Filters the queryset to include only active items.

        Returns
        -------
        AccountModelQuerySet
            A filtered queryset containing only the items marked as active.
        """
        return self.filter(active=True)

    def inactive(self) -> 'AccountModelQuerySet':
        """
        Filters and returns queryset entries where the active field is set to False.

        Returns
        -------
        AccountModelQuerySet
            A queryset containing entries with active=False.
        """
        return self.filter(active=False)

    def locked(self) -> 'AccountModelQuerySet':
        """
        Filters the queryset to include only locked AccountModels.

        Returns
        -------
        AccountModelQuerySet
            A queryset containing only the objects with locked set to True.
        """
        return self.filter(locked=True)

    def unlocked(self) -> 'AccountModelQuerySet':
        """
        Returns a filtered list of items where the 'locked' attribute is set to False.

        Returns
        -------
        AccountModelQuerySet
            A queryset of items with 'locked' attribute set to False
        """
        return self.filter(locked=False)

    def with_roles(self, roles: Union[List, str]) -> 'AccountModelQuerySet':
        """
        Filter the accounts based on the specified roles. This method helps to retrieve accounts associated
        with a particular role or a list of roles.

        For example, to get all accounts categorized under the role "asset_ppe_build" (which might include
        fixed assets like Buildings), you can utilize this method.

        Parameters
        ----------
        roles : Union[List[str], str]
            The role or a list of roles to filter the accounts by. If a single string is provided, it is converted
            into a list containing that role.

        Returns
        -------
        AccountModelQuerySet
            A QuerySet of accounts filtered by the provided roles.
        """
        roles = validate_roles(roles)
        if isinstance(roles, str):
            roles = [roles]
        roles = validate_roles(roles)
        return self.filter(role__in=roles)

    def with_codes(self, codes: Union[List, str]) -> 'AccountModelQuerySet':
        if isinstance(codes, str):
            codes = [codes]
        return self.filter(code__in=codes)

    def cash(self) -> 'AccountModelQuerySet':
        """Retrieve accounts that are of type ASSET_CA_CASH."""
        return self.filter(role__exact=ASSET_CA_CASH)

    def expenses(self) -> 'AccountModelQuerySet':
        """
        Retrieve a queryset containing expenses filtered by specified roles.

        This method filters the expenses based on roles defined in the
        `GROUP_EXPENSES` constant. It ensures that only the relevant expenses
        associated with the specified roles are included in the queryset.

        Returns
        -------
            AccountModelQuerySet
                A queryset consisting of expenses filtered according to the roles in `GROUP_EXPENSES`.
        """
        return self.filter(role__in=GROUP_EXPENSES)

    def is_coa_root(self) -> 'AccountModelQuerySet':
        """
        Retrieves the Chart of Accounts (CoA) root node queryset.

        A Chart of Accounts Root is a foundational element indicating the primary node in the
        account hierarchy. This method filters the queryset to include only the Chart of Accounts (CoA)
        root node.

        Returns
        -------
            AccountModelQuerySet
        """
        return self.filter(role__in=ROOT_GROUP)

    def not_coa_root(self) -> 'AccountModelQuerySet':
        """
        Exclude AccountModels with ROOT_GROUP role from the QuerySet.

        Returns
        -------
        AccountModelQuerySet
            A QuerySet excluding users with role in ROOT_GROUP.
        """
        return self.exclude(role__in=ROOT_GROUP)

    def gb_bs_role(self) -> 'AccountModelQuerySet':
        """
        Groups accounts by Balance Sheet Bucket and then further groups them by role.

        Returns
        -------
            List[Tuple]
                A list where each element is a tuple. The first element of the tuple is the BS bucket,
                and the second element is a list of tuples where each sub-tuple contains a role display
                and a list of accounts that fall into that role within the BS bucket.
        """
        accounts_gb = list(
            (r, sorted(list(gb), key=lambda acc: ROLES_ORDER_ALL.index(acc.role))) for r, gb in
            groupby(self, key=lambda acc: acc.get_bs_bucket())
        )
        return [
            (bsr, [
                (r, sorted(list(l), key=lambda acc: acc.code)) for r, l in
                groupby(gb, key=lambda a: a.get_role_display())
            ]) for bsr, gb in accounts_gb
        ]

    def is_role_default(self) -> 'AccountModelQuerySet':
        """
        Filter the queryset to include only entries where `role_default`
        is set to True, excluding entries marked as 'coa_root'.

        Returns
        -------
        AccountModelQuerySet
            Filtered queryset with `role_default` set to True and excluding 'coa_root' entries.
        """
        return self.not_coa_root().filter(role_default=True)

    def can_transact(self) -> 'AccountModelQuerySet':
        """
        Filter the queryset to include only accounts that can accept new transactions.

        Returns
        -------
        QuerySet
            A QuerySet containing the filtered results.
        """
        return self.filter(
            Q(locked=False) &
            Q(active=True) &
            Q(coa_model__active=True)
        )

    def available(self) -> 'AccountModelQuerySet':
        return self.filter(
            Q(locked=False) &
            Q(active=True) &
            Q(coa_model__active=True)
        )

    def for_bill(self) -> 'AccountModelQuerySet':
        """
        Retrieves only available and unlocked AccountModels for a specific EntityModel,
        specifically for the creation and management of Bills. Roles within the 'GROUP_BILL'
        context include: ASSET_CA_CASH, ASSET_CA_PREPAID, and LIABILITY_CL_ACC_PAYABLE.

        Returns
        -------
        AccountModelQuerySet
            A QuerySet of the requested EntityModel's chart of accounts.
        """
        return self.available().filter(role__in=GROUP_BILL)

    def for_invoice(self) -> 'AccountModelQuerySet':
        """
        Retrieves available and unlocked AccountModels for a specific EntityModel, specifically for the creation
        and management of Invoices.

        This method ensures that only relevant accounts are pulled, as defined under the roles in `GROUP_INVOICE`.
        These roles include: ASSET_CA_CASH, ASSET_CA_RECEIVABLES, and LIABILITY_CL_DEFERRED_REVENUE.

        Returns
        -------
        AccountModelQuerySet
            A QuerySet containing the AccountModels relevant for the specified EntityModel and the roles defined
            in `GROUP_INVOICE`.
        """
        return self.available().filter(role__in=GROUP_INVOICE)

    def for_user(self, user_model) -> 'AccountModelQuerySet':
        """
        Parameters
        ----------
        user_model : UserModel
            The user model instance to use for filtering.

        Returns
        -------
        AccountModelQuerySet
            The filtered queryset based on the user's permissions. Superusers get the complete queryset whereas other
            users get a filtered queryset based on their role as admin or manager in the entity.
        """
        if user_model.is_superuser:
            return self

        return self.filter(
            Q(coa_model__entity__admin=user_model) |
            Q(coa_model__entity__managers__in=[user_model])
        )


class AccountModelManager(MP_NodeManager):
    """
    AccountModelManager class provides methods to manage and retrieve AccountModel objects.
    It inherits from MP_NodeManager for tree-like model implementation.
    """

    def get_queryset(self) -> AccountModelQuerySet:
        """
        Retrieve and return the default AccountModel QuerySet.

        The query set is ordered by the 'path' field and uses 'select_related' to reduce the number of database queries
        by retrieving the related 'coa_model'.

        Returns
        -------
        AccountModelQuerySet
            An instance of AccountModelQuerySet ordered by 'path' and prefetching related 'coa_model'.
        """
        return AccountModelQuerySet(
            self.model,
            using=self._db
        ).order_by('path').select_related(
            'coa_model'
        ).annotate(
            _coa_slug=F('coa_model__slug'),
            _coa_active=F('coa_model__active'),
            _entity_slug=F('coa_model__entity__slug'),
        )

    @deprecated_entity_slug_behavior
    def for_entity(
            self,
            entity_model: Union['EntityModel | str | UUID'] = None,
            coa_model: Optional['ChartOfAccountModel | str | UUID'] = None,
            **kwargs
    ) -> AccountModelQuerySet:
        """
        Filters the queryset for an entity and, optionally, a chart of account (COA) model.

        The method refines the queryset based on the provided `entity_model` and, optionally,
        the `coa_model`. If a deprecated `user_model` is specified in keyword arguments,
        a warning is issued. The method supports `EntityModel`, `str`, and `UUID` types
        for both `entity_model` and `coa_model`. A validation error is raised for unsupported types.

        Parameters
        ----------
        entity_model : Union['EntityModel', str, UUID]
            The entity model used for filtering the queryset. Could be an instance of
            `EntityModel`, a string (slug), or a UUID.
        coa_model : Optional[Union['ChartOfAccountModel', str, UUID]], optional
            The COA model used for filtering the queryset. Can be an instance of
            `ChartOfAccountModel`, a string (slug), or a UUID. If None, default Entity ChartOfAccounts is used.
            Defaults to None.
        **kwargs : dict
            Additional keyword arguments. A deprecated argument `user_model` can be passed
            for backward compatibility.

        Returns
        -------
        AccountModelQuerySet
            A queryset filtered by the input entity model and, optionally, the chart of
            account model.

        Raises
        ------
        AccountModelValidationError
            If an invalid type is passed for either `entity_model` or `coa_model`.

        Warns
        -----
        DeprecationWarning
            If the `user_model` parameter is passed in the keyword arguments and the
            application relies on deprecated behavior.
        """
        EntityModel = lazy_loader.get_entity_model()
        ChartOfAccountModel = lazy_loader.get_coa_model()

        qs = self.get_queryset()

        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(coa_model__entity=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(coa_model__entity__slug__exact=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(coa_model__entity_id=entity_model)
        else:
            raise AccountModelValidationError(
                message='Must pass an instance of EntityModel, String or UUID for entity_model.'
            )

        if coa_model:
            if isinstance(coa_model, ChartOfAccountModel):
                qs = qs.filter(coa_model=coa_model)
            elif isinstance(coa_model, str):
                qs = qs.filter(coa_model__slug__exact=coa_model)
            elif isinstance(coa_model, UUID):
                qs = qs.filter(coa_model__uuid__exact=coa_model)
            else:
                raise AccountModelValidationError(
                    message='Must pass an instance of ChartOfAccountModel, String or UUID for coa_model.'
                )
            return qs

        return qs.filter(coa_model__slug__exact=F('coa_model__entity__default_coa__slug'))


def account_code_validator(value: str):
    if not value.isalnum():
        raise AccountModelValidationError(_('Account code must be alpha numeric, got {%s}') % value)


class AccountModelAbstract(MP_Node, CreateUpdateMixIn):
    """
    Abstract class representing an Account Model.

    Attributes
    ----------
    BALANCE_TYPE : list
        List of choices for the balance type of the account. Options include 'Credit' and 'Debit'.
    uuid : UUIDField
        Unique identifier for each account instance.
    code : CharField
        Code representing the account, constrained by length and specific validation rules.
    name : CharField
        Name of the account, constrained by length.
    role : CharField
        Role associated with the account, with specific predefined choices.
    role_default : BooleanField
        Flag indicating if this account is the default for its role.
    balance_type : CharField
        Type of balance the account holds, must be either 'debit' or 'credit'.
    locked : BooleanField
        Indicates whether the account is locked.
    active : BooleanField
        Indicates whether the account is active.
    coa_model : ForeignKey
        Reference to the associated ChartOfAccountModel.
    """

    BALANCE_TYPE = [
        (CREDIT, _('Credit')),
        (DEBIT, _('Debit'))
    ]

    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    code = models.CharField(max_length=10, verbose_name=_('Account Code'), validators=[account_code_validator])
    name = models.CharField(max_length=100, verbose_name=_('Account Name'))
    role = models.CharField(max_length=30, choices=ACCOUNT_ROLE_CHOICES, verbose_name=_('Account Role'))
    role_default = models.BooleanField(null=True, blank=True, verbose_name=_('Coa Role Default Account'))
    balance_type = models.CharField(max_length=6, choices=BALANCE_TYPE, verbose_name=_('Account Balance Type'))
    locked = models.BooleanField(default=False, verbose_name=_('Locked'))
    active = models.BooleanField(default=False, verbose_name=_('Active'))
    coa_model = models.ForeignKey('django_ledger.ChartOfAccountModel',
                                  on_delete=models.CASCADE,
                                  verbose_name=_('Chart of Accounts'))
    objects = AccountModelManager.from_queryset(queryset_class=AccountModelQuerySet)()

    class Meta:
        abstract = True
        ordering = ['-created']
        verbose_name = _('Account')
        verbose_name_plural = _('Accounts')
        constraints = [
            UniqueConstraint(
                fields=('coa_model', 'code'),
                name='unique_code_for_coa_model',
                violation_error_message=_('Account codes must be unique for each Chart of Accounts Model.')
            ),
            UniqueConstraint(
                fields=('coa_model', 'role', 'role_default'),
                name='only_one_account_assigned_as_default_for_role',
                violation_error_message=_('Only one default account for role permitted.')
            )
        ]
        indexes = [
            models.Index(fields=['role']),
            models.Index(fields=['balance_type']),
            models.Index(fields=['active']),
            models.Index(fields=['locked']),
            models.Index(fields=['coa_model', 'code']),
            models.Index(fields=['code'])
        ]

    def __str__(self):
        return '{x1} - {x5}: {x2} ({x3}/{x4})'.format(
            x1=self.role_bs.upper(),
            x2=self.name,
            x3=self.role.upper(),
            x4=self.balance_type,
            x5=self.code
        )

    def alt_str(self):
        """
        Returns a formatted string representation of the object.

        The formatted string includes the code, name, role, and balance type
        of the object. The role is converted to uppercase for consistency,
        and the balance type is displayed as is. This method provides a
        concise textual representation for quick identification or display.

        Returns:
            str: A formatted string in the format 'code: name (ROLE/BALANCE_TYPE)'.
        """
        return f'{self.code}: {self.name} ({self.role.upper()}/{self.balance_type})'

    @property
    def coa_slug(self):
        """
        Property that retrieves the `coa_slug` attribute from the object. If the attribute
        is not found, it fetches the `slug` attribute from the `coa_model`.

        Attributes:
            _coa_slug (str): Cached value of the `coa_slug` if it exists.
            coa_model (Any): Object containing the `slug` attribute that serves
                as a fallback when `_coa_slug` is not present.

        Returns:
            str: The value of `_coa_slug` if defined, or the `slug` attribute from
            `coa_model` if `_coa_slug` is not available.
        """
        try:
            return getattr(self, '_coa_slug')
        except AttributeError:
            return self.coa_model.slug

    @property
    def entity_slug(self):
        """
            Retrieve the slug value associated with the entity.

            This property method returns the value of the private attribute
            '_entity_slug' for the current instance. The purpose of the
            slug is typically to provide a URL-friendly string representing
            the entity.

            Returns
            -------
            Any
                The value of the '_entity_slug' attribute.
        """
        return getattr(self, '_entity_slug')

    @classmethod
    def create_account(cls,
                       name: str,
                       role: bool,
                       balance_type: str,
                       is_role_default: bool = False,
                       locked: bool = False,
                       active: bool = False,
                       **kwargs):
        """
        Create a new AccountModel instance, managing parent/child relationships properly.

        This convenience method ensures correct creation of new accounts, handling the intricate logic needed for
        maintaining hierarchical relationships between accounts.

        Parameters
        ----------
        name : str
            Name of the new account entity.
        role : str
            Role assigned to the account.
        balance_type : str
            Type of balance associated with the account. Must be either 'debit' or 'credit'.
        is_role_default : bool, optional
            Indicates if the account should be the default for its role. Only one default account per role is allowed.
            Defaults to False.
        locked : bool, optional
            Flags the account as locked. Defaults to False.
        active : bool, optional
            Flags the account as active. Defaults to True.
        **kwargs : dict, optional
            Additional attributes for account creation.

        Returns
        -------
        AccountModel
            The newly created `AccountModel` instance.
        """
        account_model = cls(
            name=name,
            role=role,
            balance_type=balance_type,
            role_default=is_role_default,
            locked=locked,
            active=active,
            **kwargs
        )
        account_model.clean()
        account_model = cls.add_root(instance=account_model)
        return account_model

    @property
    def role_bs(self) -> str:
        """
        Returns the principal role of the account on the balance sheet.

        The principal role can be one of the following:
        - 'asset'
        - 'liability'
        - 'equity'

        Returns
        -------
        str
            A string representing the principal role of the account on the balance sheet.
        """
        return BS_ROLES.get(self.role)

    def is_root_account(self):
        """
        Checks if the current user's role belongs to the ROOT_GROUP.

        Returns
        -------
        bool
            True if the role is in the ROOT_GROUP, False otherwise
        """
        return self.role in ROOT_GROUP

    def is_debit(self) -> bool:
        """
        Checks if the account has a DEBIT balance type.

        Returns
        -------
        bool
            True if account has a DEBIT balance, else False.
        """
        return self.balance_type == DEBIT

    def is_credit(self):
        """
        Checks if the Account Model has a CREDIT balance type.

        Returns
        -------
        bool
            True if account has a CREDIT balance, else False.
        """
        return self.balance_type == CREDIT

    def is_coa_root(self):
        """
        Check if the current Account Model role is 'ROOT_COA'.

        Returns
        -------
        bool
            True if the role is 'ROOT_COA', False otherwise.
        """
        return self.role == ROOT_COA

    def is_asset(self) -> bool:
        """
        Determines if the current Account Model role of the instance is considered an asset.

        Returns
        -------
        bool
            True if the role is part of the GROUP_ASSETS, False otherwise.
        """
        return self.role in GROUP_ASSETS

    def is_liability(self) -> bool:
        """
        Determines if the current Account Model role is considered a liability.

        Returns
        -------
        bool
            True if the role is part of GROUP_LIABILITIES, otherwise False.
        """
        return self.role in GROUP_LIABILITIES

    def is_capital(self) -> bool:
        """
        Checks if the current Account Model role is in the capital group.

        Returns
        -------
        bool
            True if the role is in GROUP_CAPITAL, otherwise False.
        """
        return self.role in GROUP_CAPITAL

    def is_income(self) -> bool:
        """
        Determines whether the current Account Model role belongs to the income group.

        Parameters
        ----------
        self : object
            The instance of the class containing attribute 'role'.

        Returns
        -------
        bool
            True if the role is in the GROUP_INCOME list, False otherwise.
        """
        return self.role in GROUP_INCOME

    def is_cogs(self) -> bool:
        """
        Determines if the role of the object is part of the GROUP_COGS.

        Returns
        -------
        bool
            True if the object's role is part of the GROUP_COGS, False otherwise.
        """
        return self.role in GROUP_COGS

    def is_expense(self) -> bool:
        """
        Checks if the current Account Model `role` is categorized under `GROUP_EXPENSES`.

        Parameters
        ----------
        None

        Returns
        -------
        bool
            True if `role` is in `GROUP_EXPENSES`, otherwise False.
        """
        return self.role in GROUP_EXPENSES

    def is_active(self) -> bool:
        """
        Determines if the current instance is active.

        Returns
        -------
        bool
            True if the instance is active, otherwise False
        """
        return self.active is True

    def is_locked(self) -> bool:
        """
        Determines if the current object is locked.

        Returns
        -------
        bool
            True if the object is locked, False otherwise.

        """
        return self.locked is True

    def is_coa_active(self) -> bool:
        try:
            return getattr(self, '_coa_active')
        except AttributeError:
            pass
        return self.coa_model.active

    def can_activate(self):
        """
        Determines if the object can be activated.

        Returns
        -------
        bool
            True if the object is inactive, otherwise False.
        """
        return all([
            self.active is False
        ])

    def can_deactivate(self):
        """
        Determine if the object can be deactivated.

        Checks if the `active` attribute is set to `True`.

        Returns
        -------
        bool
            True if the object is currently active and can be deactivated, otherwise False.
        """
        return all([
            self.active is True
        ])

    def can_lock(self):
        return all([
            self.locked is False
        ])

    def can_unlock(self):
        return all([
            self.locked is True
        ])

    def lock(self, commit: bool = True, raise_exception: bool = True, **kwargs):
        if not self.can_lock():
            if raise_exception:
                raise AccountModelValidationError(
                    message=_(f'Cannot lock account {self.code}: {self.name}. Active: {self.is_active()}')
                )
            return

        self.locked = True
        if commit:
            self.save(update_fields=[
                'locked',
                'updated'
            ])

    def unlock(self, commit: bool = True, raise_exception: bool = True, **kwargs):
        if not self.can_unlock():
            if raise_exception:
                raise AccountModelValidationError(
                    message=_(f'Cannot unlock account {self.code}: {self.name}. Active: {self.is_active()}')
                )
            return

        self.locked = False
        if commit:
            self.save(update_fields=[
                'locked',
                'updated'
            ])

    def activate(self, commit: bool = True, raise_exception: bool = True, **kwargs):
        """
        Checks if the Account Model instance can be activated, then Activates the AccountModel instance.
        Raises exception if AccountModel cannot be activated.

        Parameters
        ----------
        commit : bool, optional
            If True, commit the changes to the database by calling the save method.
        raise_exception : bool, optional
            If True, raises an AccountModelValidationError if the account cannot be activated.
        kwargs : dict
            Additional parameters that can be passed for further customization.
        """
        if not self.can_activate():
            if raise_exception:
                raise AccountModelValidationError(
                    message=_(f'Cannot activate account {self.code}: {self.name}. Active: {self.is_active()}')
                )
            return
        self.active = True
        if commit:
            self.save(update_fields=[
                'active',
                'updated'
            ])

    def deactivate(self, commit: bool = True, raise_exception: bool = True, **kwargs):
        """
        Checks if the Account Model instance can be de-activated, then De-activates the AccountModel instance.
        Raises exception if AccountModel cannot be de-activated.

        Parameters
        ----------
        commit : bool, optional
            If True, commit the changes to the database by calling the save method.
        raise_exception : bool, optional
            If True, raises an AccountModelValidationError if the account cannot be activated.
        kwargs : dict
            Additional parameters that can be passed for further customization.
        """
        if not self.can_deactivate():
            if raise_exception:
                raise AccountModelValidationError(
                    message=_(f'Cannot deactivate account {self.code}: {self.name}. Active: {self.is_active()}')
                )
            return
        self.active = False
        if commit:
            self.save(
                update_fields=[
                    'active',
                    'updated'
                ])

    def can_transact(self) -> bool:
        """
        Determines if a transaction can be performed based on multiple conditions.

        Returns
        -------
        bool
            True if all conditions are met, enabling a transaction; False otherwise.

        Conditions:
        1. The chart of accounts (coa_model) must be active.
        2. The entity must not be locked.
        3. The entity itself must be active.
        """
        return all([
            self.is_coa_active(),
            not self.is_locked()
        ])

    def get_code_prefix(self) -> str:
        """
        Returns the code prefix based on the account type.

        This method determines the account type by calling the respective
        account type methods and returns the corresponding code prefix based on Accounting best practices..

        Returns
        -------
        str
            The code prefix for the account type. The possible values are:
            '1' for assets, '2' for liabilities, '3' for capital,
            '4' for income, '5' for cost of goods sold (COGS),
            '6' for expenses.

        Raises
        ------
        AccountModelValidationError
            If the account role does not match any of the predefined categories.
        """
        if self.is_asset():
            return '1'
        elif self.is_liability():
            return '2'
        elif self.is_capital():
            return '3'
        elif self.is_income():
            return '4'
        elif self.is_cogs():
            return '5'
        elif self.is_expense():
            return '6'
        raise AccountModelValidationError(f'Invalid role match for role {self.role}...')

    def get_root_role(self) -> str:
        """
        Returns the root role corresponding to the account type.

        Returns
        -------
        str
            The root role corresponding to the account type.

        Raises
        ------
        AccountModelValidationError
            If no valid role match is found for the account's role.
        """
        if self.is_asset():
            return ROOT_ASSETS
        elif self.is_liability():
            return ROOT_LIABILITIES
        elif self.is_capital():
            return ROOT_CAPITAL
        elif self.is_income():
            return ROOT_INCOME
        elif self.is_cogs():
            return ROOT_GROUP
        elif self.is_expense():
            return ROOT_EXPENSES
        elif self.is_coa_root():
            return ROOT_COA
        raise AccountModelValidationError(f'Invalid role match for role {self.role}...')

    def get_account_move_choice_queryset(self):
        """
        Retrieves a filtered queryset of account models that the current Account Model instance
        can be a child of.

        The queryset is filtered based on the specified role and its hierarchical parent roles.
        Account models with a UUID matching the current instance's UUID are excluded from the results.

        Returns
        -------
        QuerySet
            A filtered set of account models suitable for moving the current instance under.
        """
        return self.coa_model.accountmodel_set.filter(
            role__in=[
                self.role,
                self.get_root_role(),
                *VALID_PARENTS.get(self.role, [])
            ],
        ).exclude(uuid__exact=self.uuid)

    def get_bs_bucket(self) -> str:
        return BS_BUCKETS[self.get_code_prefix()]

    def is_indented(self):
        """
        Check if the current depth level is greater than 2.

        Returns
        -------
        bool
            True if the depth is greater than 2, False otherwise.
        """
        return self.depth > 2

    def get_html_pixel_indent(self):
        """
        Calculates the pixel indentation for HTML elements based on the depth attribute for UI purposes

        Returns
        -------
        str
            The calculated pixel indentation as a string with 'px' suffix.
        """
        return f'{(self.depth - 2) * 40}px'

    def generate_random_code(self):
        """
        Generates a random code for the account adding a prefix 1-6 depending on account role.

        Raises
        ------
        AccountModelValidationError
            If the account role is not assigned before code generation.

        Returns
        -------
        str
            A randomly generated code prefixed with a role-based prefix.
        """
        if not self.role:
            raise AccountModelValidationError('Must assign account role before generate random code')

        prefix = self.get_code_prefix()
        ri = randint(10000, 99999)
        return f'{prefix}{ri}'

    # URLS...
    def get_absolute_url(self) -> str:
        return reverse(
            viewname='django_ledger:account-detail',
            kwargs={
                'account_pk': self.uuid,
                'entity_slug': self.entity_slug,
                'coa_slug': self.coa_slug
            }
        )

    def get_update_url(self) -> str:
        return reverse(
            viewname='django_ledger:account-update',
            kwargs={
                'account_pk': self.uuid,
                'entity_slug': self.entity_slug,
                'coa_slug': self.coa_slug
            }
        )

    def get_action_deactivate_url(self) -> str:
        return reverse(
            viewname='django_ledger:account-action-deactivate',
            kwargs={
                'account_pk': self.uuid,
                'entity_slug': self.entity_slug,
                'coa_slug': self.coa_slug
            }
        )

    def get_action_activate_url(self) -> str:
        return reverse(
            viewname='django_ledger:account-action-activate',
            kwargs={
                'account_pk': self.uuid,
                'entity_slug': self.entity_slug,
                'coa_slug': self.coa_slug
            }
        )

    def get_action_lock_url(self) -> str:
        return reverse(
            viewname='django_ledger:account-action-lock',
            kwargs={
                'account_pk': self.uuid,
                'entity_slug': self.entity_slug,
                'coa_slug': self.coa_slug
            }
        )

    def get_action_unlock_url(self) -> str:
        return reverse(
            viewname='django_ledger:account-action-unlock',
            kwargs={
                'account_pk': self.uuid,
                'entity_slug': self.entity_slug,
                'coa_slug': self.coa_slug
            }
        )

    def get_coa_account_list_url(self) -> str:
        return reverse(
            viewname='django_ledger:account-list',
            kwargs={
                'entity_slug': self.entity_slug,
                'coa_slug': self.coa_slug
            })

    def clean(self):
        if not self.code and DJANGO_LEDGER_ACCOUNT_CODE_GENERATE:
            self.code = self.generate_random_code()

        if DJANGO_LEDGER_ACCOUNT_CODE_USE_PREFIX:
            pf = self.get_code_prefix()
            if self.code[0] != pf:
                raise AccountModelValidationError(f'Account {self.get_role_display()} code {self.code} '
                                                  f'must start with {pf} for CoA consistency')


class AccountModel(AccountModelAbstract):
    """
    Base Account Model from Account Model Abstract Class
    """

    class Meta(AccountModelAbstract.Meta):
        abstract = False


def accountmodel_presave(instance: AccountModel, **kwargs):
    if instance.role_default is False:
        instance.role_default = None


pre_save.connect(receiver=accountmodel_presave, sender=AccountModel)
